@@ -4,10 +4,15 @@
  * Distributed under terms of the GPL3 license.
  */
 
+// Need _POSIX_C_SOURCE for strtok_r
+#define _POSIX_C_SOURCE 200809L
+
 #include "data-types.h"
 #include "control-codes.h"
 #include "screen.h"
 #include "graphics.h"
+#include "charsets.h"
+#include "monotonic.h"
 #include <time.h>
 
 extern PyTypeObject Screen_Type;
@@ -17,10 +22,14 @@ static uint64_t pow10_array[] = {
     1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000, 1000000000, 10000000000
 };
 
-static inline uint64_t
-utoi(uint32_t *buf, unsigned int sz) {
-    uint64_t ans = 0;
-    uint32_t *p = buf;
+static inline int64_t
+utoi(const uint32_t *buf, unsigned int sz) {
+    int64_t ans = 0;
+    const uint32_t *p = buf;
+    int mult = 1;
+    if (sz && *p == '-') {
+        mult = -1; p++; sz--;
+    }
     // Ignore leading zeros
     while(sz > 0) {
         if (*p == '0') { p++; sz--; }
@@ -31,7 +40,7 @@ utoi(uint32_t *buf, unsigned int sz) {
             ans += (p[i] - '0') * pow10_array[j];
         }
     }
-    return ans;
+    return ans * mult;
 }
 
 
@@ -47,7 +56,6 @@ utf8(char_type codepoint) {
 // }}}
 
 // Macros {{{
-#define MAX_PARAMS 256
 #define IS_DIGIT \
     case '0': \
     case '1': \
@@ -75,12 +83,12 @@ _report_error(PyObject *dump_callback, const char *fmt, ...) {
 }
 
 static void
-_report_params(PyObject *dump_callback, const char *name, unsigned int *params, unsigned int count, Region *r) {
+_report_params(PyObject *dump_callback, const char *name, int *params, unsigned int count, Region *r) {
     static char buf[MAX_PARAMS*3] = {0};
     unsigned int i, p=0;
     if (r) p += snprintf(buf + p, sizeof(buf) - 2, "%u %u %u %u ", r->top, r->left, r->bottom, r->right);
     for(i = 0; i < count && p < MAX_PARAMS*3-20; i++) {
-        int n = snprintf(buf + p, MAX_PARAMS*3 - p, "%u ", params[i]);
+        int n = snprintf(buf + p, MAX_PARAMS*3 - p, "%i ", params[i]);
         if (n < 0) break;
         p += n;
     }
@@ -119,6 +127,9 @@ _report_params(PyObject *dump_callback, const char *name, unsigned int *params,
 #define REPORT_OSC2(name, code, string) \
     Py_XDECREF(PyObject_CallFunction(dump_callback, "sIO", #name, code, string)); PyErr_Clear();
 
+#define REPORT_HYPERLINK(id, url) \
+    Py_XDECREF(PyObject_CallFunction(dump_callback, "szz", "set_active_hyperlink", id, url)); PyErr_Clear();
+
 #else
 
 #define DUMP_UNUSED UNUSED
@@ -132,6 +143,7 @@ _report_params(PyObject *dump_callback, const char *name, unsigned int *params,
 #define FLUSH_DRAW
 #define REPORT_OSC(name, string)
 #define REPORT_OSC2(name, code, string)
+#define REPORT_HYPERLINK(id, url)
 
 #endif
 
@@ -184,7 +196,7 @@ handle_normal_mode_char(Screen *screen, uint32_t ch, PyObject DUMP_UNUSED *dump_
             break;  // no-op
         default:
             REPORT_DRAW(ch);
-            screen_draw(screen, ch);
+            screen_draw(screen, ch, true);
             break;
     }
 #undef CALL_SCREEN_HANDLER
@@ -223,9 +235,9 @@ handle_esc_mode_char(Screen *screen, uint32_t ch, PyObject DUMP_UNUSED *dump_cal
                     CALL_ED(screen_save_cursor); break;
                 case ESC_DECRC:
                     CALL_ED(screen_restore_cursor); break;
-                case ESC_DECPNM:
+                case ESC_DECKPNM:
                     CALL_ED(screen_normal_keypad_mode); break;
-                case ESC_DECPAM:
+                case ESC_DECKPAM:
                     CALL_ED(screen_alternate_keypad_mode); break;
                 case '%':
                 case '(':
@@ -299,79 +311,154 @@ handle_esc_mode_char(Screen *screen, uint32_t ch, PyObject DUMP_UNUSED *dump_cal
 } // }}}
 
 // OSC mode {{{
+
+static inline bool
+parse_osc_8(char *buf, char **id, char **url) {
+    char *boundary = strstr(buf, ";");
+    if (boundary == NULL) return false;
+    *boundary = 0;
+    if (*(boundary + 1)) *url = boundary + 1;
+    char *save = NULL, *token = strtok_r(buf, ":", &save);
+    while (token != NULL) {
+        size_t len = strlen(token);
+        if (len > 3 && token[0] == 'i' && token[1] == 'd' && token[2] == '=' && token[3]) {
+            *id = token + 3;
+            break;
+        }
+        token = strtok_r(NULL, ":", &save);
+    }
+    return true;
+}
+
+static inline void
+dispatch_hyperlink(Screen *screen, size_t pos, size_t size, PyObject DUMP_UNUSED *dump_callback) {
+    // since the spec says only ASCII printable chars are allowed in OSC 8, we
+    // can just convert to char* directly
+    if (!size) return;  // ignore empty OSC 8 since it must have two semi-colons to be valid, which means one semi-colon here
+    char *id = NULL, *url = NULL;
+    char *data = malloc(size + 1);
+    if (!data) fatal("Out of memory");
+    for (size_t i = 0; i < size; i++) {
+        data[i] = screen->parser_buf[i + pos] & 0x7f;
+        if (data[i] < 32 || data[i] > 126) data[i] = '_';
+    }
+    data[size] = 0;
+
+    if (parse_osc_8(data, &id, &url)) {
+        REPORT_HYPERLINK(id, url);
+        set_active_hyperlink(screen, id, url);
+    } else {
+        REPORT_ERROR("Ignoring malformed OSC 8 code");
+    }
+
+    free(data);
+}
+
 static inline void
 dispatch_osc(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
+#define DISPATCH_OSC_WITH_CODE(name) REPORT_OSC2(name, code, string); name(screen, code, string);
 #define DISPATCH_OSC(name) REPORT_OSC(name, string); name(screen, string);
-#define SET_COLOR(name) REPORT_OSC2(name, code, string); name(screen, code, string);
+#define START_DISPATCH {\
+    PyObject *string = PyUnicode_FromKindAndData(PyUnicode_4BYTE_KIND, screen->parser_buf + i, limit - i); \
+    if (string) {
+#define END_DISPATCH Py_CLEAR(string); } PyErr_Clear(); break; }
+
     const unsigned int limit = screen->parser_buf_pos;
-    unsigned int code=0, i;
-    for (i = 0; i < MIN(limit, 5); i++) {
+    int code=0;
+    unsigned int i;
+    for (i = 0; i < MIN(limit, 5u); i++) {
         if (screen->parser_buf[i] < '0' || screen->parser_buf[i] > '9') break;
     }
     if (i > 0) {
         code = utoi(screen->parser_buf, i);
         if (i < limit && screen->parser_buf[i] == ';') i++;
     }
-    PyObject *string = PyUnicode_FromKindAndData(PyUnicode_4BYTE_KIND, screen->parser_buf + i, limit - i);
-    if (string != NULL) {
-        switch(code) {
-            case 0:
-                DISPATCH_OSC(set_title);
-                DISPATCH_OSC(set_icon);
-                break;
-            case 1:
-                DISPATCH_OSC(set_icon);
-                break;
-            case 2:
-                DISPATCH_OSC(set_title);
-                break;
-            case 4:
-            case 104:
-                SET_COLOR(set_color_table_color);
-                break;
-            case 10:
-            case 11:
-            case 12:
-            case 17:
-            case 19:
-            case 110:
-            case 111:
-            case 112:
-            case 117:
-            case 119:
-                SET_COLOR(set_dynamic_color);
-                break;
-            case 52:
-                DISPATCH_OSC(clipboard_control);
-                break;
-            case 30001:
-                REPORT_COMMAND(screen_push_dynamic_colors);
-                screen_push_dynamic_colors(screen);
-                break;
-            case 30101:
-                REPORT_COMMAND(screen_pop_dynamic_colors);
-                screen_pop_dynamic_colors(screen);
-                break;
-            default:
-                REPORT_ERROR("Unknown OSC code: %u", code);
-                break;
-        }
-        Py_CLEAR(string);
+    switch(code) {
+        case 0:
+            START_DISPATCH
+            DISPATCH_OSC(set_title);
+            DISPATCH_OSC(set_icon);
+            END_DISPATCH
+        case 1:
+            START_DISPATCH
+            DISPATCH_OSC(set_icon);
+            END_DISPATCH
+        case 2:
+            START_DISPATCH
+            DISPATCH_OSC(set_title);
+            END_DISPATCH
+        case 4:
+        case 104:
+            START_DISPATCH
+            DISPATCH_OSC_WITH_CODE(set_color_table_color);
+            END_DISPATCH
+        case 7:
+            START_DISPATCH
+            DISPATCH_OSC_WITH_CODE(process_cwd_notification);
+            END_DISPATCH
+        case 8:
+            dispatch_hyperlink(screen, i, limit-i, dump_callback);
+            break;
+        case 9:
+        case 99:
+            START_DISPATCH
+            DISPATCH_OSC_WITH_CODE(desktop_notify)
+            END_DISPATCH
+        case 10:
+        case 11:
+        case 12:
+        case 17:
+        case 19:
+        case 110:
+        case 111:
+        case 112:
+        case 117:
+        case 119:
+            START_DISPATCH
+            DISPATCH_OSC_WITH_CODE(set_dynamic_color);
+            END_DISPATCH
+        case 52:
+            START_DISPATCH
+            DISPATCH_OSC(clipboard_control);
+            END_DISPATCH
+        case 30001:
+            REPORT_COMMAND(screen_push_dynamic_colors);
+            screen_push_colors(screen, 0);
+            break;
+        case 30101:
+            REPORT_COMMAND(screen_pop_dynamic_colors);
+            screen_pop_colors(screen, 0);
+            break;
+        default:
+            REPORT_ERROR("Unknown OSC code: %u", code);
+            break;
     }
 #undef DISPATCH_OSC
-#undef SET_COLOR
+#undef DISPATCH_OSC_WITH_CODE
+#undef START_DISPATCH
+#undef END_DISPATCH
 }
 // }}}
 
 // CSI mode {{{
+// As per ECMA 48 section 5.4 secondary byte is column 02 of the 7-bit ascii table
 #define CSI_SECONDARY \
-        case ';': \
-        case ':': \
-        case '"': \
-        case '*': \
-        case '\'': \
         case ' ': \
-        case '$':
+        case '!': \
+        case '"': \
+        case '#': \
+        case '$': \
+        case '%': \
+        case '&': \
+        case '\'': \
+        case '(': \
+        case ')': \
+        case '*': \
+        case '+': \
+        case ',': \
+        case '-': \
+        case '.': \
+        case '/':
 
 
 static inline void
@@ -379,16 +466,16 @@ screen_cursor_up2(Screen *s, unsigned int count) { screen_cursor_up(s, count, fa
 static inline void
 screen_cursor_back1(Screen *s, unsigned int count) { screen_cursor_back(s, count, -1); }
 static inline void
-screen_tabn(Screen *s, unsigned int count) { for (index_type i=0; i < MAX(1, count); i++) screen_tab(s); }
+screen_tabn(Screen *s, unsigned int count) { for (index_type i=0; i < MAX(1u, count); i++) screen_tab(s); }
 
 static inline const char*
-repr_csi_params(unsigned int *params, unsigned int num_params) {
+repr_csi_params(int *params, unsigned int num_params) {
     if (!num_params) return "";
     static char buf[256];
-    unsigned int pos = 0;
-    while (pos < 200 && num_params && sizeof(buf) > pos + 1) {
-        const char *fmt = num_params > 1 ? "%u " : "%u";
-        int ret = snprintf(buf + pos, sizeof(buf) - pos - 1, fmt, params[num_params--]);
+    unsigned int pos = 0, i = 0;
+    while (pos < 200 && i++ < num_params && sizeof(buf) > pos + 1) {
+        const char *fmt = i < num_params ? "%i, " : "%i";
+        int ret = snprintf(buf + pos, sizeof(buf) - pos - 1, fmt, params[i-1]);
         if (ret < 0) return "An error occurred formatting the params array";
         pos += ret;
     }
@@ -396,8 +483,11 @@ repr_csi_params(unsigned int *params, unsigned int num_params) {
     return buf;
 }
 
-static inline void
-parse_sgr(Screen *screen, uint32_t *buf, unsigned int num, unsigned int *params, PyObject DUMP_UNUSED *dump_callback, const char *report_name DUMP_UNUSED, Region *region) {
+#ifdef DUMP_COMMANDS
+static
+#endif
+void
+parse_sgr(Screen *screen, uint32_t *buf, unsigned int num, int *params, PyObject DUMP_UNUSED *dump_callback, const char *report_name DUMP_UNUSED, Region *region) {
     enum State { START, NORMAL, MULTIPLE, COLOR, COLOR1, COLOR3 };
     enum State state = START;
     unsigned int num_params, num_start, i;
@@ -525,7 +615,8 @@ parse_sgr(Screen *screen, uint32_t *buf, unsigned int num, unsigned int *params,
 
 static inline unsigned int
 parse_region(Region *r, uint32_t *buf, unsigned int num) {
-    unsigned int i, start, params[8] = {0}, num_params=0;
+    unsigned int i, start, num_params = 0;
+    int params[8] = {0};
     for (i=0, start=0; i < num && num_params < 4; i++) {
         switch(buf[i]) {
             IS_DIGIT
@@ -559,34 +650,61 @@ parse_region(Region *r, uint32_t *buf, unsigned int num) {
 
 static inline void
 dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
+#define AT_MOST_ONE_PARAMETER { \
+    if (num_params > 1) { \
+        REPORT_ERROR("CSI code 0x%x has %u > 1 parameters", code, num_params); \
+        break; \
+    } \
+}
+#define NON_NEGATIVE_PARAM(x) { \
+    if (x < 0) { \
+        REPORT_ERROR("CSI code 0x%x is not allowed to have negative parameter (%d)", code, x); \
+        break; \
+    } \
+}
+
 #define CALL_CSI_HANDLER1(name, defval) \
+    AT_MOST_ONE_PARAMETER; \
     p1 = num_params > 0 ? params[0] : defval; \
+    NON_NEGATIVE_PARAM(p1); \
     REPORT_COMMAND(name, p1); \
     name(screen, p1); \
     break;
 
 #define CALL_CSI_HANDLER1P(name, defval, qch) \
+    AT_MOST_ONE_PARAMETER; \
     p1 = num_params > 0 ? params[0] : defval; \
+    NON_NEGATIVE_PARAM(p1); \
     private = start_modifier == qch; \
     REPORT_COMMAND(name, p1, private); \
     name(screen, p1, private); \
     break;
 
 #define CALL_CSI_HANDLER1S(name, defval) \
+    AT_MOST_ONE_PARAMETER; \
     p1 = num_params > 0 ? params[0] : defval; \
+    NON_NEGATIVE_PARAM(p1); \
     REPORT_COMMAND(name, p1, start_modifier); \
     name(screen, p1, start_modifier); \
     break;
 
 #define CALL_CSI_HANDLER1M(name, defval) \
+    AT_MOST_ONE_PARAMETER; \
     p1 = num_params > 0 ? params[0] : defval; \
+    NON_NEGATIVE_PARAM(p1); \
     REPORT_COMMAND(name, p1, end_modifier); \
     name(screen, p1, end_modifier); \
     break;
 
 #define CALL_CSI_HANDLER2(name, defval1, defval2) \
+    if (num_params > 2) { \
+        REPORT_ERROR("CSI code 0x%x has %u > 2 parameters", code, num_params); \
+        break; \
+    } \
     p1 = num_params > 0 ? params[0] : defval1; \
     p2 = num_params > 1 ? params[1] : defval2; \
+    NON_NEGATIVE_PARAM(p1); \
+    NON_NEGATIVE_PARAM(p2); \
     REPORT_COMMAND(name, p1, p2); \
     name(screen, p1, p2); \
     break;
@@ -594,43 +712,59 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
 #define SET_MODE(func) \
     p1 = start_modifier == '?' ? 5 : 0; \
     for (i = 0; i < num_params; i++) { \
+        NON_NEGATIVE_PARAM(params[i]); \
         REPORT_COMMAND(func, params[i], start_modifier == '?'); \
         func(screen, params[i] << p1); \
     } \
     break;
 
+#define NO_MODIFIERS(modifier, special, special_msg) { \
+    if (start_modifier || end_modifier) { \
+        if (special && modifier == special) { REPORT_ERROR(special_msg); } \
+        else { REPORT_ERROR("CSI code 0x%x has unsupported start modifier: 0x%x or end modifier: 0x%x", code, start_modifier, end_modifier);} \
+        break; \
+    } \
+}
+
     char start_modifier = 0, end_modifier = 0;
     uint32_t *buf = screen->parser_buf, code = screen->parser_buf[screen->parser_buf_pos];
-    unsigned int num = screen->parser_buf_pos, start, i, num_params=0, p1, p2;
-    static unsigned int params[MAX_PARAMS] = {0};
+    unsigned int num = screen->parser_buf_pos, start, i, num_params=0;
+    static int params[MAX_PARAMS] = {0}, p1, p2;
     bool private;
-    if (buf[0] == '>' || buf[0] == '?' || buf[0] == '!') {
+    if (buf[0] == '>' || buf[0] == '<' || buf[0] == '?' || buf[0] == '!' || buf[0] == '=') {
         start_modifier = (char)screen->parser_buf[0];
         buf++; num--;
     }
-    if (code == SGR && !start_modifier) {
+    if (num > 0) {
+        switch(buf[num-1]) {
+            CSI_SECONDARY
+                end_modifier = (char)buf[--num];
+                break;
+        }
+    }
+    if (code == SGR && !start_modifier && !end_modifier) {
         parse_sgr(screen, buf, num, params, dump_callback, "select_graphic_rendition", NULL);
         return;
     }
-    if (code == 'r' && !start_modifier && num > 0 && buf[num - 1] == '$') {
+    if (code == 'r' && !start_modifier && end_modifier == '$') {
         // DECCARA
         Region r = {0};
-        unsigned int consumed = parse_region(&r, buf, --num);
+        unsigned int consumed = parse_region(&r, buf, num);
         num -= consumed; buf += consumed;
         parse_sgr(screen, buf, num, params, dump_callback, "deccara", &r);
         return;
     }
 
-    if (num > 0) {
-        switch(buf[num-1]) {
-            CSI_SECONDARY
-                end_modifier = (char)buf[--num];
-        }
-    }
     for (i=0, start=0; i < num; i++) {
         switch(buf[i]) {
             IS_DIGIT
                 break;
+            case '-':
+                if (i > start) {
+                    REPORT_ERROR("CSI code can contain hyphens only at the start of numbers");
+                    return;
+                }
+                break;
             default:
                 if (i > start) params[num_params++] = utoi(buf + start, i - start);
                 else if (i == start && buf[i] == ';') params[num_params++] = 0;
@@ -641,8 +775,12 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
     if (i > start) params[num_params++] = utoi(buf + start, i - start);
     switch(code) {
         case ICH:
+            NO_MODIFIERS(end_modifier, ' ', "Shift left escape code not implemented");
             CALL_CSI_HANDLER1(screen_insert_characters, 1);
+        case REP:
+            CALL_CSI_HANDLER1(screen_repeat_character, 1);
         case CUU:
+            NO_MODIFIERS(end_modifier, ' ', "Shift right escape code not implemented");
             CALL_CSI_HANDLER1(screen_cursor_up2, 1);
         case CUD:
         case VPR:
@@ -677,7 +815,23 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
         case DL:
             CALL_CSI_HANDLER1(screen_delete_lines, 1);
         case DCH:
-            CALL_CSI_HANDLER1(screen_delete_characters, 1);
+            if (end_modifier == '#' && !start_modifier) {
+                CALL_CSI_HANDLER1(screen_push_colors, 0);
+            } else {
+                CALL_CSI_HANDLER1(screen_delete_characters, 1);
+            }
+        case 'Q':
+            if (end_modifier == '#' && !start_modifier) { CALL_CSI_HANDLER1(screen_pop_colors, 0); }
+            REPORT_ERROR("Unknown CSI Q sequence with start and end modifiers: '%c' '%c' and %u parameters", start_modifier, end_modifier, num_params);
+            break;
+        case 'R':
+            if (end_modifier == '#' && !start_modifier) {
+                REPORT_COMMAND(screen_report_color_stack);
+                screen_report_color_stack(screen);
+                break;
+            }
+            REPORT_ERROR("Unknown CSI R sequence with start and end modifiers: '%c' '%c' and %u parameters", start_modifier, end_modifier, num_params);
+            break;
         case ECH:
             CALL_CSI_HANDLER1(screen_erase_characters, 1);
         case DA:
@@ -708,7 +862,7 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
                 break;
             }
             if (start_modifier || end_modifier) {
-                REPORT_ERROR("Unknown CSI t sequence with start and end modifiers: '%c' '%c', %u parameters and first parameter: %u", start_modifier, end_modifier, num_params, params[0]);
+                REPORT_ERROR("Unknown CSI t sequence with start and end modifiers: '%c' '%c', %u parameters and first parameter: %d", start_modifier, end_modifier, num_params, params[0]);
                 break;
             }
             switch(params[0]) {
@@ -726,7 +880,7 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
                     CALL_CSI_HANDLER2(screen_manipulate_title_stack, 22, 0);
                     break;
                 default:
-                    REPORT_ERROR("Unknown CSI t window manipulation sequence with %u parameters and first parameter: %u", num_params, params[0]);
+                    REPORT_ERROR("Unknown CSI t window manipulation sequence with %u parameters and first parameter: %d", num_params, params[0]);
                     break;
             }
             break;
@@ -736,6 +890,23 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
                 screen_restore_cursor(screen);
                 break;
             }
+            if (!end_modifier && start_modifier == '?') {
+                REPORT_COMMAND(screen_report_key_encoding_flags);
+                screen_report_key_encoding_flags(screen);
+                break;
+            }
+            if (!end_modifier && start_modifier == '=') {
+                CALL_CSI_HANDLER2(screen_set_key_encoding_flags, 0, 1);
+                break;
+            }
+            if (!end_modifier && start_modifier == '>') {
+                CALL_CSI_HANDLER1(screen_push_key_encoding_flags, 0);
+                break;
+            }
+            if (!end_modifier && start_modifier == '<') {
+                CALL_CSI_HANDLER1(screen_pop_key_encoding_flags, 1);
+                break;
+            }
             REPORT_ERROR("Unknown CSI u sequence with start and end modifiers: '%c' '%c' and %u parameters", start_modifier, end_modifier, num_params);
             break;
         case 'r':
@@ -758,6 +929,7 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
         case DECSCUSR:
             CALL_CSI_HANDLER1M(screen_set_cursor, 1);
         case SU:
+            NO_MODIFIERS(end_modifier, ' ', "Select presentation directions escape code not implemented");
             CALL_CSI_HANDLER1(screen_scroll, 1);
         case SD:
             CALL_CSI_HANDLER1(screen_reverse_scroll, 1);
@@ -769,6 +941,12 @@ dispatch_csi(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
                 REPORT_ERROR("Unknown DECSTR CSI sequence with start and end modifiers: '%c' '%c'", start_modifier, end_modifier);
             }
             break;
+        case 'm':
+            if (start_modifier == '>' && !end_modifier) {
+                CALL_CSI_HANDLER2(screen_xtmodkeys, 0, 0);
+                break;
+            }
+            /* fallthrough */
         default:
             REPORT_ERROR("Unknown CSI code: '%s' with start_modifier: '%c' and end_modifier: '%c' and parameters: '%s'", utf8(code), start_modifier, end_modifier, repr_csi_params(params, num_params));
     }
@@ -807,12 +985,14 @@ dispatch_dcs(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
             }
             break;
         case PENDING_MODE_CHAR:
-            if (screen->parser_buf_pos > 2 && (screen->parser_buf[1] == '1' || screen->parser_buf[2] == '2') && screen->parser_buf[2] == 's') {
+            if (screen->parser_buf_pos > 2 && (screen->parser_buf[1] == '1' || screen->parser_buf[1] == '2') && screen->parser_buf[2] == 's') {
                 if (screen->parser_buf[1] == '1') {
                     screen->pending_mode.activated_at = monotonic();
                     REPORT_COMMAND(screen_start_pending_mode);
                 } else {
-                    // ignore stop without matching start
+                    // ignore stop without matching start, see _queue_pending_bytes()
+                    // for how stop is detected while in pending mode.
+                    REPORT_ERROR("Pending mode stop command issued while not in pending mode");
                     REPORT_COMMAND(screen_stop_pending_mode);
                 }
             } else {
@@ -831,7 +1011,8 @@ dispatch_dcs(Screen *screen, PyObject DUMP_UNUSED *dump_callback) {
 #undef CMD_PREFIX
 #define PRINT_PREFIX "kitty-print|"
             } else if (startswith(screen->parser_buf + 1, screen->parser_buf_pos - 1, PRINT_PREFIX)) {
-                PyObject *msg = PyUnicode_FromKindAndData(PyUnicode_4BYTE_KIND, screen->parser_buf + sizeof(PRINT_PREFIX), screen->parser_buf_pos - sizeof(PRINT_PREFIX));
+                const size_t pp_size = sizeof(PRINT_PREFIX);
+                PyObject *msg = PyUnicode_FromKindAndData(PyUnicode_4BYTE_KIND, screen->parser_buf + pp_size, screen->parser_buf_pos - pp_size);
                 if (msg != NULL) {
                     REPORT_OSC2(screen_handle_print, (char)screen->parser_buf[0], msg);
                     screen_handle_print(screen, msg);
@@ -946,6 +1127,9 @@ accumulate_oth(Screen *screen, uint32_t ch, PyObject DUMP_UNUSED *dump_callback)
     switch(ch) {
         case ST:
             return true;
+        case DEL:
+        case NUL:
+            break;
         case ESC_ST:
             if (screen->parser_buf_pos > 0 && screen->parser_buf[screen->parser_buf_pos - 1] == ESC) {
                 screen->parser_buf_pos--;
@@ -976,12 +1160,15 @@ accumulate_csi(Screen *screen, uint32_t ch, PyObject DUMP_UNUSED *dump_callback)
     switch(ch) {
         IS_DIGIT
         CSI_SECONDARY
+        case ':':
+        case ';':
             ENSURE_SPACE;
             screen->parser_buf[screen->parser_buf_pos++] = ch;
             break;
         case '?':
         case '>':
-        case '!':
+        case '<':
+        case '=':
             if (screen->parser_buf_pos != 0) {
                 REPORT_ERROR("Invalid character in CSI: 0x%x, ignoring the sequence", ch);
                 SET_STATE(0);
@@ -1183,7 +1370,7 @@ end:
 }
 
 static inline void
-do_parse_bytes(Screen *screen, const uint8_t *read_buf, const size_t read_buf_sz, double now, PyObject *dump_callback DUMP_UNUSED) {
+do_parse_bytes(Screen *screen, const uint8_t *read_buf, const size_t read_buf_sz, monotonic_t now, PyObject *dump_callback DUMP_UNUSED) {
     enum STATE {START, PARSE_PENDING, PARSE_READ_BUF, QUEUE_PENDING};
     enum STATE state = START;
     size_t read_buf_pos = 0;
@@ -1259,7 +1446,7 @@ FNAME(parse_bytes)(PyObject UNUSED *self, PyObject *args) {
 
 
 void
-FNAME(parse_worker)(Screen *screen, PyObject *dump_callback, double now) {
+FNAME(parse_worker)(Screen *screen, PyObject *dump_callback, monotonic_t now) {
 #ifdef DUMP_COMMANDS
     if (screen->read_buf_sz) {
         Py_XDECREF(PyObject_CallFunction(dump_callback, "sy#", "bytes", screen->read_buf, screen->read_buf_sz)); PyErr_Clear();
