@@ -4,30 +4,31 @@
 
 import shlex
 import sys
-from typing import Generator, List, NamedTuple, Optional, Tuple, Union
+from typing import Generator, List, Optional, Sequence, Union
 
 from .cli_stub import CLIOptions
 from .config_data import to_layout_names
-from .constants import FloatEdges, kitty_exe
-from .layout import all_layouts
+from .constants import kitty_exe
+from .layout.interface import all_layouts
 from .options_stub import Options
+from .os_window_size import WindowSize, WindowSizeData, WindowSizes
 from .typing import SpecialWindowInstance
 from .utils import log_error, resolved_shell
 
 
-class WindowSizeOpts(NamedTuple):
-
-    initial_window_width: Tuple[int, str]
-    initial_window_height: Tuple[int, str]
-    window_margin_width: FloatEdges
-    window_padding_width: FloatEdges
-    remember_window_size: bool
+def get_os_window_sizing_data(opts: Options, session: Optional['Session'] = None) -> WindowSizeData:
+    if session is None or session.os_window_size is None:
+        sizes = WindowSizes(WindowSize(*opts.initial_window_width), WindowSize(*opts.initial_window_height))
+    else:
+        sizes = session.os_window_size
+    return WindowSizeData(sizes, opts.remember_window_size, opts.single_window_margin_width, opts.window_margin_width, opts.window_padding_width)
 
 
 class Tab:
 
     def __init__(self, opts: Options, name: str):
-        self.windows: List[Union[List[str], 'SpecialWindowInstance']] = []
+        from .launch import LaunchSpec
+        self.windows: List[Union[LaunchSpec, 'SpecialWindowInstance']] = []
         self.name = name.strip()
         self.active_window_idx = 0
         self.enabled_layouts = opts.enabled_layouts
@@ -38,11 +39,13 @@ class Tab:
 
 class Session:
 
-    def __init__(self, default_title: Optional[str] = None):
+    def __init__(self, default_title: Optional[str] = None, default_watchers: Sequence[str] = ()):
         self.tabs: List[Tab] = []
+        self.default_watchers = list(default_watchers)
         self.active_tab_idx = 0
         self.default_title = default_title
-        self.os_window_size: Optional[WindowSizeOpts] = None
+        self.os_window_size: Optional[WindowSizes] = None
+        self.os_window_class: Optional[str] = None
 
     def add_tab(self, opts: Options, name: str = '') -> None:
         if self.tabs and not self.tabs[-1].windows:
@@ -58,13 +61,15 @@ class Session:
         self.tabs[-1].layout = val
 
     def add_window(self, cmd: Union[None, str, List[str]]) -> None:
-        if cmd:
-            cmd = shlex.split(cmd) if isinstance(cmd, str) else cmd
-        else:
-            cmd = None
-        from .tabs import SpecialWindow
+        from .launch import parse_launch_args
+        if isinstance(cmd, str):
+            cmd = shlex.split(cmd)
+        spec = parse_launch_args(cmd)
+        if self.default_watchers:
+            spec.opts.watcher = list(spec.opts.watcher) + self.default_watchers
         t = self.tabs[-1]
-        t.windows.append(SpecialWindow(cmd, cwd=t.cwd, override_title=t.next_title or self.default_title))
+        spec.opts.cwd = spec.opts.cwd or t.cwd
+        t.windows.append(spec)
         t.next_title = None
 
     def add_special_window(self, sw: 'SpecialWindowInstance') -> None:
@@ -86,9 +91,10 @@ class Session:
 def parse_session(raw: str, opts: Options, default_title: Optional[str] = None) -> Generator[Session, None, None]:
 
     def finalize_session(ans: Session) -> Session:
+        from .tabs import SpecialWindow
         for t in ans.tabs:
             if not t.windows:
-                t.windows.append(resolved_shell(opts))
+                t.windows.append(SpecialWindow(cmd=resolved_shell(opts), override_title=default_title))
         return ans
 
     ans = Session(default_title)
@@ -123,7 +129,9 @@ def parse_session(raw: str, opts: Options, default_title: Optional[str] = None)
             elif cmd == 'os_window_size':
                 from kitty.config_data import window_size
                 w, h = map(window_size, rest.split(maxsplit=1))
-                ans.os_window_size = WindowSizeOpts(w, h, opts.window_margin_width, opts.window_padding_width, False)
+                ans.os_window_size = WindowSizes(WindowSize(*w), WindowSize(*h))
+            elif cmd == 'os_window_class':
+                ans.os_window_class = rest
             else:
                 raise ValueError('Unknown command in session file: {}'.format(cmd))
     yield finalize_session(ans)
@@ -155,7 +163,8 @@ def create_sessions(
         else:
             yield from parse_session(session_data, opts, getattr(args, 'title', None))
             return
-    ans = Session()
+    default_watchers = args.watcher if args else ()
+    ans = Session(default_watchers=default_watchers)
     current_layout = opts.enabled_layouts[0] if opts.enabled_layouts else 'tall'
     ans.add_tab(opts)
     ans.tabs[-1].layout = current_layout
