@@ -6,6 +6,7 @@
  */
 
 #include "state.h"
+#include "cleanup.h"
 #include <math.h>
 
 GlobalState global_state = {{0}};
@@ -60,6 +61,27 @@ GlobalState global_state = {{0}};
             if (wp->id == cb_window_id && cb_window_id) global_state.callback_os_window = wp; \
     }}
 
+static inline double
+dpi_for_os_window_id(id_type os_window_id) {
+    double dpi = 0;
+    if (os_window_id) {
+        WITH_OS_WINDOW(os_window_id)
+            dpi = (os_window->logical_dpi_x + os_window->logical_dpi_y) / 2.;
+        END_WITH_OS_WINDOW
+    }
+    if (dpi == 0) {
+        dpi = (global_state.default_dpi.x + global_state.default_dpi.y) / 2.;
+    }
+    return dpi;
+}
+
+
+static long
+pt_to_px(double pt, id_type os_window_id) {
+    const double dpi = dpi_for_os_window_id(os_window_id);
+    return ((long)round((pt * (dpi / 72.0))));
+}
+
 
 OSWindow*
 current_os_window() {
@@ -84,6 +106,20 @@ os_window_for_kitty_window(id_type kitty_window_id) {
     return NULL;
 }
 
+Window*
+window_for_window_id(id_type kitty_window_id) {
+    for (size_t i = 0; i < global_state.num_os_windows; i++) {
+        OSWindow *w = global_state.os_windows + i;
+        for (size_t t = 0; t < w->num_tabs; t++) {
+            Tab *tab = w->tabs + t;
+            for (size_t c = 0; c < tab->num_windows; c++) {
+                if (tab->windows[c].id == kitty_window_id) return tab->windows + c;
+            }
+        }
+    }
+    return NULL;
+}
+
 static void
 send_bgimage_to_gpu(BackgroundImageLayout layout, BackgroundImage *bgimage) {
     RepeatStrategy r;
@@ -243,7 +279,16 @@ destroy_window(Window *w) {
 
 static inline void
 remove_window_inner(Tab *tab, id_type id) {
+    id_type active_window_id = 0;
+    if (tab->active_window < tab->num_windows) active_window_id = tab->windows[tab->active_window].id;
     REMOVER(tab->windows, id, tab->num_windows, destroy_window, tab->capacity);
+    if (active_window_id) {
+        for (unsigned int w = 0; w < tab->num_windows; w++) {
+            if (tab->windows[w].id == active_window_id) {
+                tab->active_window = w; break;
+            }
+        }
+    }
 }
 
 static inline void
@@ -329,8 +374,17 @@ destroy_tab(Tab *tab) {
 
 static inline void
 remove_tab_inner(OSWindow *os_window, id_type id) {
+    id_type active_tab_id = 0;
+    if (os_window->active_tab < os_window->num_tabs) active_tab_id = os_window->tabs[os_window->active_tab].id;
     make_os_window_context_current(os_window);
     REMOVER(os_window->tabs, id, os_window->num_tabs, destroy_tab, os_window->capacity);
+    if (active_tab_id) {
+        for (unsigned int i = 0; i < os_window->num_tabs; i++) {
+            if (os_window->tabs[i].id == active_tab_id) {
+                os_window->active_tab = i; break;
+            }
+        }
+    }
 }
 
 static inline void
@@ -382,11 +436,12 @@ set_active_tab(id_type os_window_id, unsigned int idx) {
 }
 
 static inline void
-set_active_window(id_type os_window_id, id_type tab_id, unsigned int idx) {
-    WITH_TAB(os_window_id, tab_id)
-        tab->active_window = idx;
+set_active_window(id_type os_window_id, id_type tab_id, id_type window_id) {
+    WITH_WINDOW(os_window_id, tab_id, window_id)
+        (void)window;
+        tab->active_window = w;
         osw->needs_render = true;
-    END_WITH_TAB;
+    END_WITH_WINDOW;
 }
 
 static inline void
@@ -398,15 +453,6 @@ swap_tabs(id_type os_window_id, unsigned int a, unsigned int b) {
     END_WITH_OS_WINDOW
 }
 
-static inline void
-swap_windows(id_type os_window_id, id_type tab_id, unsigned int a, unsigned int b) {
-    WITH_TAB(os_window_id, tab_id);
-    Window w = tab->windows[b];
-    tab->windows[b] = tab->windows[a];
-    tab->windows[a] = w;
-    END_WITH_TAB;
-}
-
 static void
 add_borders_rect(id_type os_window_id, id_type tab_id, uint32_t left, uint32_t top, uint32_t right, uint32_t bottom, uint32_t color) {
     WITH_TAB(os_window_id, tab_id)
@@ -450,8 +496,35 @@ mark_os_window_for_close(OSWindow* w, CloseRequest cr) {
     w->close_request = cr;
 }
 
+static bool
+owners_for_window_id(id_type window_id, OSWindow **os_window, Tab **tab) {
+    if (os_window) *os_window = NULL;
+    if (tab) *tab = NULL;
+    for (size_t o = 0; o < global_state.num_os_windows; o++) {
+        OSWindow *osw = global_state.os_windows + o;
+        for (size_t t = 0; t < osw->num_tabs; t++) {
+            Tab *qtab = osw->tabs + t;
+            for (size_t w = 0; w < qtab->num_windows; w++) {
+                Window *window = qtab->windows + w;
+                if (window->id == window_id) {
+                    if (os_window) *os_window = osw;
+                    if (tab) *tab = qtab;
+                    return true;
+    }}}}
+    return false;
+}
 
 
+bool
+make_window_context_current(id_type window_id) {
+    OSWindow *os_window;
+    if (owners_for_window_id(window_id, &os_window, NULL)) {
+        make_os_window_context_current(os_window);
+        return true;
+    }
+    return false;
+}
+
 
 // Python API {{{
 #define PYWRAP0(name) static PyObject* py##name(PYNOARG)
@@ -466,7 +539,9 @@ mark_os_window_for_close(OSWindow* w, CloseRequest cr) {
 #define KI(name) PYWRAP1(name) { id_type a; unsigned int b; PA("KI", &a, &b); name(a, b); Py_RETURN_NONE; }
 #define KII(name) PYWRAP1(name) { id_type a; unsigned int b, c; PA("KII", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
 #define KKI(name) PYWRAP1(name) { id_type a, b; unsigned int c; PA("KKI", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
+#define KKK(name) PYWRAP1(name) { id_type a, b, c; PA("KKK", &a, &b, &c); name(a, b, c); Py_RETURN_NONE; }
 #define KKII(name) PYWRAP1(name) { id_type a, b; unsigned int c, d; PA("KKII", &a, &b, &c, &d); name(a, b, c, d); Py_RETURN_NONE; }
+#define KKKK(name) PYWRAP1(name) { id_type a, b, c, d; PA("KKKK", &a, &b, &c, &d); name(a, b, c, d); Py_RETURN_NONE; }
 #define KK5I(name) PYWRAP1(name) { id_type a, b; unsigned int c, d, e, f, g; PA("KKIIIII", &a, &b, &c, &d, &e, &f, &g); name(a, b, c, d, e, f, g); Py_RETURN_NONE; }
 #define BOOL_SET(name) PYWRAP1(set_##name) { global_state.name = PyObject_IsTrue(args); Py_RETURN_NONE; }
 
@@ -587,17 +662,6 @@ set_url_prefixes(PyObject *up) {
     PyObject *key, *value; Py_ssize_t pos = 0; \
     while (PyDict_Next(d, &pos, &key, &value))
 
-static inline void
-set_special_keys(PyObject *dict) {
-    dict_iter(dict) {
-        if (!PyTuple_Check(key)) { PyErr_SetString(PyExc_TypeError, "dict keys for special keys must be tuples"); return; }
-        int mods = PyLong_AsLong(PyTuple_GET_ITEM(key, 0));
-        bool is_native = PyTuple_GET_ITEM(key, 1) == Py_True;
-        int glfw_key = PyLong_AsLong(PyTuple_GET_ITEM(key, 2));
-        set_special_key_combo(glfw_key, mods, is_native);
-    }}
-}
-
 static inline float
 PyFloat_AsFloat(PyObject *o) {
     return (float)PyFloat_AsDouble(o);
@@ -646,6 +710,7 @@ PYWRAP1(set_options) {
     S(dynamic_background_opacity, PyObject_IsTrue);
     S(inactive_text_alpha, PyFloat_AsFloat);
     S(scrollback_pager_history_size, PyLong_AsUnsignedLong);
+    S(scrollback_fill_enlarged_window, PyObject_IsTrue);
     S(cursor_shape, PyLong_AsLong);
     S(cursor_beam_thickness, PyFloat_AsFloat);
     S(cursor_underline_thickness, PyFloat_AsFloat);
@@ -690,7 +755,11 @@ PYWRAP1(set_options) {
     S(force_ltr, PyObject_IsTrue);
     S(resize_draw_strategy, PyLong_AsLong);
     S(resize_in_steps, PyObject_IsTrue);
+    S(allow_hyperlinks, PyObject_IsTrue);
     S(pointer_shape_when_grabbed, pointer_shape);
+    S(default_pointer_shape, pointer_shape);
+    S(pointer_shape_when_dragging, pointer_shape);
+    S(detect_urls, PyObject_IsTrue);
 
     GA(tab_bar_style);
     global_state.tab_bar_hidden = PyUnicode_CompareWithASCIIString(ret, "hidden") == 0 ? true: false;
@@ -712,9 +781,9 @@ PYWRAP1(set_options) {
     OPT(select_by_word_characters_count) = PyUnicode_GET_LENGTH(chars);
     Py_DECREF(chars);
 
-    GA(keymap); set_special_keys(ret);
+    GA(keymap);
     Py_DECREF(ret); if (PyErr_Occurred()) return NULL;
-    GA(sequence_map); set_special_keys(ret);
+    GA(sequence_map);
     Py_DECREF(ret); if (PyErr_Occurred()) return NULL;
 
     GA(background_image); background_image(ret); Py_CLEAR(ret);
@@ -822,6 +891,19 @@ PYWRAP1(mark_os_window_for_close) {
     Py_RETURN_FALSE;
 }
 
+PYWRAP1(set_application_quit_request) {
+    CloseRequest cr = IMPERATIVE_CLOSE_REQUESTED;
+    PA("|i", &cr);
+    global_state.quit_request = cr;
+    global_state.has_pending_closes = true;
+    request_tick_callback();
+    Py_RETURN_NONE;
+}
+
+PYWRAP0(current_application_quit_request) {
+    return Py_BuildValue("i", global_state.quit_request);
+}
+
 PYWRAP1(focus_os_window) {
     id_type os_window_id;
     int also_raise = 1;
@@ -872,14 +954,6 @@ PYWRAP1(background_opacity_of) {
     Py_RETURN_NONE;
 }
 
-static inline bool
-fix_window_idx(Tab *tab, id_type window_id, unsigned int *window_idx) {
-    for (id_type fix = 0; fix < tab->num_windows; fix++) {
-        if (tab->windows[fix].id == window_id) { *window_idx = fix; return true; }
-    }
-    return false;
-}
-
 PYWRAP1(set_window_padding) {
     id_type os_window_id, tab_id, window_id;
     unsigned int left, top, right, bottom;
@@ -894,22 +968,18 @@ PYWRAP1(set_window_render_data) {
 #define A(name) &(d.name)
 #define B(name) &(g.name)
     id_type os_window_id, tab_id, window_id;
-    unsigned int window_idx;
     ScreenRenderData d = {0};
     WindowGeometry g = {0};
-    PA("KKKIffffOIIII", &os_window_id, &tab_id, &window_id, &window_idx, A(xstart), A(ystart), A(dx), A(dy), A(screen), B(left), B(top), B(right), B(bottom));
+    PA("KKKffffOIIII", &os_window_id, &tab_id, &window_id, A(xstart), A(ystart), A(dx), A(dy), A(screen), B(left), B(top), B(right), B(bottom));
 
-    WITH_TAB(os_window_id, tab_id);
-        if (tab->windows[window_idx].id != window_id) {
-            if (!fix_window_idx(tab, window_id, &window_idx)) Py_RETURN_NONE;
-        }
-        Py_CLEAR(tab->windows[window_idx].render_data.screen);
-        d.vao_idx = tab->windows[window_idx].render_data.vao_idx;
-        d.gvao_idx = tab->windows[window_idx].render_data.gvao_idx;
-        tab->windows[window_idx].render_data = d;
-        tab->windows[window_idx].geometry = g;
-        Py_INCREF(tab->windows[window_idx].render_data.screen);
-    END_WITH_TAB;
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+        Py_CLEAR(window->render_data.screen);
+        d.vao_idx = window->render_data.vao_idx;
+        d.gvao_idx = window->render_data.gvao_idx;
+        window->render_data = d;
+        window->geometry = g;
+        Py_INCREF(window->render_data.screen);
+    END_WITH_WINDOW;
     Py_RETURN_NONE;
 #undef A
 #undef B
@@ -917,15 +987,13 @@ PYWRAP1(set_window_render_data) {
 
 PYWRAP1(update_window_visibility) {
     id_type os_window_id, tab_id, window_id;
-    unsigned int window_idx;
     int visible;
-    PA("KKKIp", &os_window_id, &tab_id, &window_id, &window_idx, &visible);
-    WITH_TAB(os_window_id, tab_id);
-        if (tab->windows[window_idx].id != window_id) {
-            if (!fix_window_idx(tab, window_id, &window_idx)) Py_RETURN_NONE;
-        }
-        tab->windows[window_idx].visible = visible & 1;
-    END_WITH_TAB;
+    PA("KKKp", &os_window_id, &tab_id, &window_id, &visible);
+    WITH_WINDOW(os_window_id, tab_id, window_id);
+        bool was_visible = window->visible & 1;
+        window->visible = visible & 1;
+        if (!was_visible && window->visible) global_state.check_for_active_animated_images = true;
+    END_WITH_WINDOW;
     Py_RETURN_NONE;
 }
 
@@ -940,26 +1008,11 @@ PYWRAP1(sync_os_window_title) {
 }
 
 
-static inline double
-dpi_for_os_window_id(id_type os_window_id) {
-    double dpi = 0;
-    if (os_window_id) {
-        WITH_OS_WINDOW(os_window_id)
-            dpi = (os_window->logical_dpi_x + os_window->logical_dpi_y) / 2.;
-        END_WITH_OS_WINDOW
-    }
-    if (dpi == 0) {
-        dpi = (global_state.default_dpi.x + global_state.default_dpi.y) / 2.;
-    }
-    return dpi;
-}
-
 PYWRAP1(pt_to_px) {
-    double pt, dpi = 0;
+    double pt;
     id_type os_window_id = 0;
     PA("d|K", &pt, &os_window_id);
-    dpi = dpi_for_os_window_id(os_window_id);
-    return PyLong_FromLong((long)round((pt * (dpi / 72.0))));
+    return PyLong_FromLong(pt_to_px(pt, os_window_id));
 }
 
 PYWRAP1(global_font_size) {
@@ -1112,9 +1165,8 @@ PYWRAP1(add_window) { PyObject *title; id_type a, b; PA("KKO", &a, &b, &title);
 PYWRAP0(current_os_window) { OSWindow *w = current_os_window(); if (!w) Py_RETURN_NONE; return PyLong_FromUnsignedLongLong(w->id); }
 TWO_ID(remove_tab)
 KI(set_active_tab)
-KKI(set_active_window)
+KKK(set_active_window)
 KII(swap_tabs)
-KKII(swap_windows)
 KK5I(add_borders_rect)
 
 #define M(name, arg_type) {#name, (PyCFunction)name, arg_type, NULL}
@@ -1138,7 +1190,6 @@ static PyMethodDef module_methods[] = {
     MW(set_active_tab, METH_VARARGS),
     MW(set_active_window, METH_VARARGS),
     MW(swap_tabs, METH_VARARGS),
-    MW(swap_windows, METH_VARARGS),
     MW(add_borders_rect, METH_VARARGS),
     MW(set_tab_bar_render_data, METH_VARARGS),
     MW(set_window_render_data, METH_VARARGS),
@@ -1147,6 +1198,8 @@ static PyMethodDef module_methods[] = {
     MW(cell_size_for_window, METH_VARARGS),
     MW(os_window_has_background_image, METH_VARARGS),
     MW(mark_os_window_for_close, METH_VARARGS),
+    MW(set_application_quit_request, METH_VARARGS),
+    MW(current_application_quit_request, METH_NOARGS),
     MW(set_titlebar_color, METH_VARARGS),
     MW(focus_os_window, METH_VARARGS),
     MW(mark_tab_bar_dirty, METH_O),
@@ -1196,10 +1249,10 @@ init_state(PyObject *module) {
     if (PyStructSequence_InitType2(&RegionType, &region_desc) != 0) return false;
     Py_INCREF((PyObject *) &RegionType);
     PyModule_AddObject(module, "Region", (PyObject *) &RegionType);
-    if (Py_AtExit(finalize) != 0) {
-        PyErr_SetString(PyExc_RuntimeError, "Failed to register the state at exit handler");
-        return false;
-    }
+    PyModule_AddIntConstant(module, "IMPERATIVE_CLOSE_REQUESTED", IMPERATIVE_CLOSE_REQUESTED);
+    PyModule_AddIntConstant(module, "NO_CLOSE_REQUESTED", NO_CLOSE_REQUESTED);
+    PyModule_AddIntConstant(module, "CLOSE_BEING_CONFIRMED", CLOSE_BEING_CONFIRMED);
+    register_at_exit_cleanup_func(STATE_CLEANUP_FUNC, finalize);
     return true;
 }
 // }}}
