@@ -32,6 +32,7 @@ typedef struct {
     unsigned int terminal_select_modifiers;
     unsigned int url_style;
     unsigned int scrollback_pager_history_size;
+    bool scrollback_fill_enlarged_window;
     char_type select_by_word_characters[256]; size_t select_by_word_characters_count;
     color_type url_color, background, foreground, active_border_color, inactive_border_color, bell_border_color;
     color_type mark1_foreground, mark1_background, mark2_foreground, mark2_background, mark3_foreground, mark3_background;
@@ -63,12 +64,16 @@ typedef struct {
     bool close_on_child_death;
     bool window_alert_on_bell;
     bool debug_keyboard;
+    bool allow_hyperlinks;
     monotonic_t resize_debounce_time;
     MouseShape pointer_shape_when_grabbed;
+    MouseShape default_pointer_shape;
+    MouseShape pointer_shape_when_dragging;
     struct {
         UrlPrefix *values;
         size_t num, max_prefix_len;
     } url_prefixes;
+    bool detect_urls;
 } Options;
 
 typedef struct {
@@ -111,6 +116,7 @@ typedef struct {
     WindowGeometry geometry;
     ClickQueue click_queue;
     monotonic_t last_drag_scroll_at;
+    uint32_t last_special_key_pressed;
 } Window;
 
 typedef struct {
@@ -131,15 +137,13 @@ typedef struct {
     BorderRects border_rects;
 } Tab;
 
-#define MAX_KEY_COUNT 512
-
 typedef struct {
     int x, y, w, h;
     bool is_set;
 } OSWindowGeometry;
 
 enum RENDER_STATE { RENDER_FRAME_NOT_REQUESTED, RENDER_FRAME_REQUESTED, RENDER_FRAME_READY };
-typedef enum { NO_CLOSE_REQUESTED, CONFIRMABLE_CLOSE_REQUESTED, IMPERATIVE_CLOSE_REQUESTED } CloseRequest;
+typedef enum { NO_CLOSE_REQUESTED, CONFIRMABLE_CLOSE_REQUESTED, CLOSE_BEING_CONFIRMED, IMPERATIVE_CLOSE_REQUESTED } CloseRequest;
 
 typedef struct {
     monotonic_t last_resize_event_at;
@@ -167,9 +171,8 @@ typedef struct {
     monotonic_t cursor_blink_zero_time, last_mouse_activity_at;
     double mouse_x, mouse_y;
     double logical_dpi_x, logical_dpi_y, font_sz_in_pts;
-    bool mouse_button_pressed[20];
+    bool mouse_button_pressed[32];
     PyObject *window_title;
-    bool is_key_pressed[MAX_KEY_COUNT];
     bool viewport_size_dirty, viewport_updated_at_least_once;
     LiveResizeInfo live_resize;
     bool has_pending_resizes, is_semi_transparent, shown_once, is_damaged;
@@ -197,16 +200,17 @@ typedef struct {
     OSWindow *os_windows;
     size_t num_os_windows, capacity;
     OSWindow *callback_os_window;
-    bool terminate;
     bool is_wayland;
     bool has_render_frames;
     bool debug_rendering, debug_font_fallback;
     bool has_pending_resizes, has_pending_closes;
     bool in_sequence_mode;
     bool tab_bar_hidden;
+    bool check_for_active_animated_images;
     double font_sz_in_pts;
     struct { double x, y; } default_dpi;
     id_type active_drag_in_window;
+    CloseRequest quit_request;
 } GlobalState;
 
 extern GlobalState global_state;
@@ -227,7 +231,7 @@ void update_os_window_viewport(OSWindow *window, bool);
 bool should_os_window_be_rendered(OSWindow* w);
 void wakeup_main_loop(void);
 void swap_window_buffers(OSWindow *w);
-void make_window_context_current(OSWindow *w);
+bool make_window_context_current(id_type);
 void hide_mouse(OSWindow *w);
 bool is_mouse_hidden(OSWindow *w);
 void destroy_os_window(OSWindow *w);
@@ -256,16 +260,22 @@ void set_titlebar_color(OSWindow *w, color_type color);
 FONTS_DATA_HANDLE load_fonts_data(double, double, double);
 void send_prerendered_sprites_for_window(OSWindow *w);
 #ifdef __APPLE__
-void get_cocoa_key_equivalent(int, int, char *key, size_t key_sz, int*);
+void get_cocoa_key_equivalent(uint32_t, int, char *key, size_t key_sz, int*);
 typedef enum {
+    NO_COCOA_PENDING_ACTION = 0,
     PREFERENCES_WINDOW = 1,
     NEW_OS_WINDOW = 2,
     NEW_OS_WINDOW_WITH_WD = 4,
-    NEW_TAB_WITH_WD = 8
+    NEW_TAB_WITH_WD = 8,
+    CLOSE_OS_WINDOW = 16,
+    CLOSE_TAB = 32,
+    NEW_TAB = 64,
+    NEXT_TAB = 128,
+    PREVIOUS_TAB = 256,
+    DETACH_TAB = 512,
+    OPEN_FILE = 1024
 } CocoaPendingAction;
 void set_cocoa_pending_action(CocoaPendingAction action, const char*);
-bool application_quit_requested(void);
-void request_application_quit(void);
 #endif
 void request_frame_render(OSWindow *w);
 void request_tick_callback(void);
@@ -279,3 +289,5 @@ void stop_main_loop(void);
 void os_window_update_size_increments(OSWindow *window);
 void set_os_window_title_from_window(Window *w, OSWindow *os_window);
 void update_os_window_title(OSWindow *os_window);
+void fake_scroll(Window *w, int amount, bool upwards);
+Window* window_for_window_id(id_type kitty_window_id);
