@@ -2,17 +2,19 @@
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
-from typing import Any, Dict, NamedTuple, Optional, Sequence, Set, Tuple
+from functools import lru_cache
+from typing import Any, Dict, NamedTuple, Optional, Sequence, Tuple
 
 from .config import build_ansi_color_table
-from .constants import WindowGeometry
 from .fast_data_types import (
     DECAWM, Screen, cell_size_for_window, pt_to_px, set_tab_bar_render_data,
     viewport_for_window
 )
-from .layout import Rect
+from .layout.base import Rect
 from .options_stub import Options
-from .rgb import Color, alpha_blend, color_from_int
+from .rgb import Color, alpha_blend, color_as_sgr, color_from_int, to_color
+from .types import WindowGeometry
+from .typing import PowerlineStyle
 from .utils import color_as_int, log_error
 from .window import calculate_gl_geometry
 
@@ -21,6 +23,9 @@ class TabBarData(NamedTuple):
     title: str
     is_active: bool
     needs_attention: bool
+    num_windows: int
+    layout_name: str
+    has_activity_since_last_focus: bool
 
 
 class DrawData(NamedTuple):
@@ -37,13 +42,54 @@ class DrawData(NamedTuple):
     default_bg: Color
     title_template: str
     active_title_template: Optional[str]
+    tab_activity_symbol: Optional[str]
+    powerline_style: PowerlineStyle
 
 
 def as_rgb(x: int) -> int:
     return (x << 8) | 2
 
 
-template_failures: Set[str] = set()
+@lru_cache()
+def report_template_failure(template: str, e: str) -> None:
+    log_error('Invalid tab title template: "{}" with error: {}'.format(template, e))
+
+
+@lru_cache()
+def compile_template(template: str) -> Any:
+    try:
+        return compile('f"""' + template + '"""', '<template>', 'eval')
+    except Exception as e:
+        report_template_failure(template, str(e))
+
+
+class ColorFormatter:
+
+    def __init__(self, which: str):
+        self.which = which
+
+    def __getattr__(self, name: str) -> str:
+        q = name
+        if q == 'default':
+            ans = '9'
+        else:
+            if name.startswith('_'):
+                q = '#' + name[1:]
+            c = to_color(q)
+            if c is None:
+                raise AttributeError(f'{name} is not a valid color')
+            ans = '8' + color_as_sgr(c)
+        return f'\x1b[{self.which}{ans}m'
+
+
+class Formatter:
+    reset = '\x1b[0m'
+    fg = ColorFormatter('3')
+    bg = ColorFormatter('4')
+    bold = '\x1b[1m'
+    nobold = '\x1b[22m'
+    italic = '\x1b[3m'
+    noitalic = '\x1b[23m'
 
 
 def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int) -> None:
@@ -52,17 +98,36 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int)
         screen.cursor.fg = draw_data.bell_fg
         screen.draw('🔔 ')
         screen.cursor.fg = fg
+    if tab.has_activity_since_last_focus and draw_data.tab_activity_symbol:
+        fg = screen.cursor.fg
+        screen.cursor.fg = draw_data.bell_fg
+        screen.draw(draw_data.tab_activity_symbol)
+        screen.cursor.fg = fg
+
     template = draw_data.title_template
     if tab.is_active and draw_data.active_title_template is not None:
         template = draw_data.active_title_template
     try:
-        title = template.format(title=tab.title, index=index)
+        eval_locals = {
+            'index': index,
+            'layout_name': tab.layout_name,
+            'num_windows': tab.num_windows,
+            'title': tab.title,
+            'fmt': Formatter,
+        }
+        title = eval(compile_template(template), {'__builtins__': {}}, eval_locals)
     except Exception as e:
-        if template not in template_failures:
-            template_failures.add(template)
-            log_error('Invalid tab title template: "{}" with error: {}'.format(template, e))
+        report_template_failure(template, str(e))
         title = tab.title
-    screen.draw(title)
+    if '\x1b' in title:
+        import re
+        for x in re.split('(\x1b\\[[^m]*m)', title):
+            if x.startswith('\x1b') and x.endswith('m'):
+                screen.apply_sgr(x[2:-1])
+            else:
+                screen.draw(x)
+    else:
+        screen.draw(title)
 
 
 def draw_tab_with_separator(draw_data: DrawData, screen: Screen, tab: TabBarData, before: int, max_title_length: int, index: int, is_last: bool) -> int:
@@ -84,7 +149,7 @@ def draw_tab_with_separator(draw_data: DrawData, screen: Screen, tab: TabBarData
     screen.cursor.fg = 0
     if not is_last:
         screen.cursor.bg = as_rgb(color_as_int(draw_data.inactive_bg))
-    screen.draw(draw_data.sep)
+        screen.draw(draw_data.sep)
     screen.cursor.bg = 0
     return end
 
@@ -117,19 +182,26 @@ def draw_tab_with_fade(draw_data: DrawData, screen: Screen, tab: TabBarData, bef
     return end
 
 
+powerline_symbols: Dict[PowerlineStyle, Tuple[str, str]] = {
+    'slanted': ('', '╱'),
+    'round': ('', '')
+}
+
+
 def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData, before: int, max_title_length: int, index: int, is_last: bool) -> int:
     tab_bg = as_rgb(color_as_int(draw_data.active_bg if tab.is_active else draw_data.inactive_bg))
     tab_fg = as_rgb(color_as_int(draw_data.active_fg if tab.is_active else draw_data.inactive_fg))
     inactive_bg = as_rgb(color_as_int(draw_data.inactive_bg))
     default_bg = as_rgb(color_as_int(draw_data.default_bg))
 
+    separator_symbol, separator_alt_symbol = powerline_symbols.get(draw_data.powerline_style, ('', ''))
     min_title_length = 1 + 2
 
     if screen.cursor.x + min_title_length >= screen.columns:
         screen.cursor.x -= 2
         screen.cursor.bg = default_bg
         screen.cursor.fg = inactive_bg
-        screen.draw('   ')
+        screen.draw(f'{separator_symbol}   ')
         return screen.cursor.x
 
     start_draw = 2
@@ -137,7 +209,7 @@ def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData
         screen.cursor.x -= 2
         screen.cursor.fg = inactive_bg
         screen.cursor.bg = tab_bg
-        screen.draw(' ')
+        screen.draw(f'{separator_symbol} ')
         screen.cursor.fg = tab_fg
     elif screen.cursor.x == 0:
         screen.cursor.bg = tab_bg
@@ -161,9 +233,9 @@ def draw_tab_with_powerline(draw_data: DrawData, screen: Screen, tab: TabBarData
             screen.cursor.bg = default_bg
         else:
             screen.cursor.bg = inactive_bg
-        screen.draw('')
+        screen.draw(separator_symbol)
     else:
-        screen.draw(' ')
+        screen.draw(f' {separator_alt_symbol}')
 
     end = screen.cursor.x
     if end < screen.columns:
@@ -209,7 +281,9 @@ class TabBar:
             self.opts.tab_fade, self.opts.active_tab_foreground, self.opts.active_tab_background,
             self.opts.inactive_tab_foreground, self.opts.inactive_tab_background,
             self.opts.tab_bar_background or self.opts.background, self.opts.tab_title_template,
-            self.opts.active_tab_title_template
+            self.opts.active_tab_title_template,
+            self.opts.tab_activity_symbol,
+            self.opts.tab_powerline_style
         )
         if self.opts.tab_bar_style == 'separator':
             self.draw_func = draw_tab_with_separator
@@ -244,7 +318,7 @@ class TabBar:
             return
         self.cell_width = cell_width
         s = self.screen
-        viewport_width = tab_bar.width - 2 * self.margin_width
+        viewport_width = max(4 * cell_width, tab_bar.width - 2 * self.margin_width)
         ncells = viewport_width // cell_width
         s.resize(1, ncells)
         s.reset_mode(DECAWM)
