@@ -2,32 +2,35 @@
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+import os
+import stat
 import weakref
 from collections import deque
 from contextlib import suppress
 from functools import partial
+from operator import attrgetter
 from typing import (
-    Deque, Dict, Generator, Iterator, List, NamedTuple,
-    Optional, Pattern, Sequence, Tuple, cast
+    Any, Deque, Dict, Generator, Iterator, List, NamedTuple, Optional, Pattern,
+    Sequence, Tuple, Union, cast
 )
 
 from .borders import Borders
 from .child import Child
 from .cli_stub import CLIOptions
-from .constants import appname, is_macos, is_wayland
+from .constants import appname, kitty_exe
 from .fast_data_types import (
     add_tab, attach_window, detach_window, get_boss, mark_tab_bar_dirty,
-    next_window_id, pt_to_px, remove_tab, remove_window, ring_bell,
-    set_active_tab, swap_tabs, x11_window_id
-)
-from .layout import (
-    Layout, Rect, create_layout_object_for, evict_cached_layouts
+    next_window_id, remove_tab, remove_window, ring_bell, set_active_tab,
+    set_active_window, swap_tabs, sync_os_window_title
 )
+from .layout.base import Layout, Rect
+from .layout.interface import create_layout_object_for, evict_cached_layouts
 from .options_stub import Options
 from .tab_bar import TabBar, TabBarData
-from .utils import log_error, resolved_shell
-from .window import Window, WindowDict, Watchers
-from .typing import TypedDict, SessionTab, SessionType
+from .typing import EdgeLiteral, SessionTab, SessionType, TypedDict
+from .utils import log_error, platform_window_id, resolved_shell
+from .window import Watchers, Window, WindowDict
+from .window_list import WindowList
 
 
 class TabDict(TypedDict):
@@ -35,6 +38,7 @@ class TabDict(TypedDict):
     is_focused: bool
     title: str
     layout: str
+    layout_state: Dict[str, Any]
     windows: List[WindowDict]
     active_window_history: List[int]
 
@@ -47,6 +51,7 @@ class SpecialWindowInstance(NamedTuple):
     cwd: Optional[str]
     overlay_for: Optional[int]
     env: Optional[Dict[str, str]]
+    watchers: Optional[Watchers]
 
 
 def SpecialWindow(
@@ -56,9 +61,10 @@ def SpecialWindow(
     cwd_from: Optional[int] = None,
     cwd: Optional[str] = None,
     overlay_for: Optional[int] = None,
-    env: Optional[Dict[str, str]] = None
+    env: Optional[Dict[str, str]] = None,
+    watchers: Optional[Watchers] = None
 ) -> SpecialWindowInstance:
-    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env)
+    return SpecialWindowInstance(cmd, stdin, override_title, cwd_from, cwd, overlay_for, env, watchers)
 
 
 def add_active_id_to_history(items: Deque[int], item_id: int, maxlen: int = 64) -> None:
@@ -79,19 +85,16 @@ class Tab:  # {{{
         cwd_from: Optional[int] = None,
         no_initial_window: bool = False
     ):
-        self._active_window_idx = 0
         self.tab_manager_ref = weakref.ref(tab_manager)
         self.os_window_id: int = tab_manager.os_window_id
         self.id: int = add_tab(self.os_window_id)
-        self.active_window_history: Deque[int] = deque()
         if not self.id:
             raise Exception('No OS window with id {} found, or tab counter has wrapped'.format(self.os_window_id))
         self.opts, self.args = tab_manager.opts, tab_manager.args
-        self.recalculate_sizes(update_layout=False)
         self.name = getattr(session_tab, 'name', '')
         self.enabled_layouts = [x.lower() for x in getattr(session_tab, 'enabled_layouts', None) or self.opts.enabled_layouts]
         self.borders = Borders(self.os_window_id, self.id, self.opts)
-        self.windows: Deque[Window] = deque()
+        self.windows = WindowList(self)
         for i, which in enumerate('first second third fourth fifth sixth seventh eighth ninth tenth'.split()):
             setattr(self, which + '_window', partial(self.nth_window, num=i))
         self._last_used_layout: Optional[str] = None
@@ -113,32 +116,17 @@ class Tab:  # {{{
             self._set_current_layout(l0)
             self.startup(session_tab)
 
-    def recalculate_sizes(self, update_layout: bool = True) -> None:
-        self.margin_width, self.padding_width, self.single_window_margin_width = map(
-            lambda x: pt_to_px(getattr(self.opts, x), self.os_window_id), (
-                'window_margin_width', 'window_padding_width', 'single_window_margin_width'))
-        self.border_width = pt_to_px(self.opts.window_border_width, self.os_window_id)
-        if update_layout and self.current_layout:
-            self.current_layout.update_sizes(
-                self.margin_width, self.single_window_margin_width, self.padding_width, self.border_width)
-
     def take_over_from(self, other_tab: 'Tab') -> None:
         self.name, self.cwd = other_tab.name, other_tab.cwd
         self.enabled_layouts = list(other_tab.enabled_layouts)
         if other_tab._current_layout_name:
             self._set_current_layout(other_tab._current_layout_name)
         self._last_used_layout = other_tab._last_used_layout
-
-        orig_windows = deque(other_tab.windows)
-        orig_history = deque(other_tab.active_window_history)
-        orig_active = other_tab._active_window_idx
         for window in other_tab.windows:
             detach_window(other_tab.os_window_id, other_tab.id, window.id)
-        other_tab.windows = deque()
-        other_tab._active_window_idx = 0
-        self.active_window_history = orig_history
-        self.windows = orig_windows
-        self._active_window_idx = orig_active
+        self.windows = other_tab.windows
+        self.windows.change_tab(self)
+        other_tab.windows = WindowList(other_tab)
         for window in self.windows:
             window.change_tab(self)
             attach_window(self.os_window_id, self.id, window.id)
@@ -148,47 +136,46 @@ class Tab:  # {{{
         self._last_used_layout = self._current_layout_name
         self.current_layout = self.create_layout_object(layout_name)
         self._current_layout_name = layout_name
+        self.mark_tab_bar_dirty()
 
     def startup(self, session_tab: 'SessionTab') -> None:
         for cmd in session_tab.windows:
-            if isinstance(cmd, (SpecialWindowInstance,)):
+            if isinstance(cmd, SpecialWindowInstance):
                 self.new_special_window(cmd)
             else:
-                self.new_window(cmd=cmd)
-        self.set_active_window_idx(session_tab.active_window_idx)
+                from .launch import launch
+                launch(get_boss(), cmd.opts, cmd.args, target_tab=self, force_target_tab=True)
+        self.windows.set_active_window_group_for(self.windows.all_windows[session_tab.active_window_idx])
 
-    @property
-    def active_window_idx(self) -> int:
-        return self._active_window_idx
+    def serialize_state(self) -> Dict[str, Any]:
+        return {
+            'version': 1,
+            'id': self.id,
+            'window_list': self.windows.serialize_state(),
+            'current_layout': self._current_layout_name,
+            'last_used_layout': self._last_used_layout,
+            'name': self.name,
+        }
 
-    @active_window_idx.setter
-    def active_window_idx(self, val: int) -> None:
-        try:
-            old_active_window: Optional[Window] = self.windows[self._active_window_idx]
-        except Exception:
-            old_active_window = None
-        else:
-            assert old_active_window is not None
-            wid = old_active_window.id if old_active_window.overlay_for is None else old_active_window.overlay_for
-            add_active_id_to_history(self.active_window_history, wid)
-        self._active_window_idx = max(0, min(val, len(self.windows) - 1))
-        try:
-            new_active_window: Optional[Window] = self.windows[self._active_window_idx]
-        except Exception:
-            new_active_window = None
-        if old_active_window is not new_active_window:
-            if old_active_window is not None:
-                old_active_window.focus_changed(False)
-            if new_active_window is not None:
-                new_active_window.focus_changed(True)
-            tm = self.tab_manager_ref()
-            if tm is not None:
-                self.relayout_borders()
-                tm.mark_tab_bar_dirty()
+    def active_window_changed(self) -> None:
+        w = self.active_window
+        set_active_window(self.os_window_id, self.id, 0 if w is None else w.id)
+        self.mark_tab_bar_dirty()
+        self.relayout_borders()
+        self.current_layout.update_visibility(self.windows)
+
+    def mark_tab_bar_dirty(self) -> None:
+        tm = self.tab_manager_ref()
+        if tm is not None:
+            tm.mark_tab_bar_dirty()
 
     @property
     def active_window(self) -> Optional[Window]:
-        return self.windows[self.active_window_idx] if self.windows else None
+        return self.windows.active_window
+
+    @property
+    def active_window_for_cwd(self) -> Optional[Window]:
+        return self.windows.active_group_base
 
     @property
     def title(self) -> str:
@@ -196,51 +183,37 @@ class Tab:  # {{{
 
     def set_title(self, title: str) -> None:
         self.name = title or ''
-        tm = self.tab_manager_ref()
-        if tm is not None:
-            tm.mark_tab_bar_dirty()
+        self.mark_tab_bar_dirty()
 
     def title_changed(self, window: Window) -> None:
         if window is self.active_window:
             tm = self.tab_manager_ref()
             if tm is not None:
-                tm.mark_tab_bar_dirty()
+                tm.title_changed(self)
 
     def on_bell(self, window: Window) -> None:
-        tm = self.tab_manager_ref()
-        if tm is not None:
-            self.relayout_borders()
-            tm.mark_tab_bar_dirty()
-
-    def visible_windows(self) -> Generator[Window, None, None]:
-        for w in self.windows:
-            if w.is_visible_in_layout:
-                yield w
+        self.mark_tab_bar_dirty()
 
     def relayout(self) -> None:
         if self.windows:
-            self.active_window_idx = self.current_layout(self.windows, self.active_window_idx)
+            self.current_layout(self.windows)
         self.relayout_borders()
 
     def relayout_borders(self) -> None:
         tm = self.tab_manager_ref()
         if tm is not None:
-            visible_windows = [w for w in self.windows if w.is_visible_in_layout]
             w = self.active_window
+            ly = self.current_layout
             self.borders(
-                windows=visible_windows, active_window=w,
-                current_layout=self.current_layout, extra_blank_rects=tm.blank_rects,
-                padding_width=self.padding_width, border_width=self.border_width,
-                draw_window_borders=self.current_layout.needs_window_borders and len(visible_windows) > 1
+                all_windows=self.windows,
+                current_layout=ly, extra_blank_rects=tm.blank_rects,
+                draw_window_borders=(ly.needs_window_borders and self.windows.num_visble_groups > 1) or ly.must_draw_borders
             )
             if w is not None:
                 w.change_titlebar_color()
 
     def create_layout_object(self, name: str) -> Layout:
-        return create_layout_object_for(
-            name, self.os_window_id, self.id, self.margin_width,
-            self.single_window_margin_width, self.padding_width,
-            self.border_width)
+        return create_layout_object_for(name, self.os_window_id, self.id)
 
     def next_layout(self) -> None:
         if len(self.enabled_layouts) > 1:
@@ -281,9 +254,9 @@ class Tab:  # {{{
             raise ValueError(increment)
         is_horizontal = quality in ('wider', 'narrower')
         increment *= 1 if quality in ('wider', 'taller') else -1
-        if self.resize_window_by(
-                self.windows[self.active_window_idx].id,
-                increment, is_horizontal) is not None:
+        w = self.active_window
+        if w is not None and self.resize_window_by(
+                w.id, increment, is_horizontal) is not None:
             ring_bell()
 
     def reset_window_sizes(self) -> None:
@@ -291,12 +264,10 @@ class Tab:  # {{{
             self.relayout()
 
     def layout_action(self, action_name: str, args: Sequence[str]) -> None:
-        ret = self.current_layout.layout_action(action_name, args, self.windows, self.active_window_idx)
+        ret = self.current_layout.layout_action(action_name, args, self.windows)
         if ret is None:
             ring_bell()
             return
-        if not isinstance(ret, bool) and isinstance(ret, int):
-            self.active_window_idx = ret
         self.relayout()
 
     def launch_child(
@@ -309,28 +280,62 @@ class Tab:  # {{{
         env: Optional[Dict[str, str]] = None,
         allow_remote_control: bool = False
     ) -> Child:
+        check_for_suitability = True
         if cmd is None:
             if use_shell:
                 cmd = resolved_shell(self.opts)
+                check_for_suitability = False
             else:
+                if self.args.args:
+                    cmd = list(self.args.args)
+                else:
+                    cmd = resolved_shell(self.opts)
+                    check_for_suitability = False
                 cmd = self.args.args or resolved_shell(self.opts)
+        if check_for_suitability:
+            old_exe = cmd[0]
+            if not os.path.isabs(old_exe):
+                import shutil
+                actual_exe = shutil.which(old_exe)
+                old_exe = actual_exe if actual_exe else os.path.abspath(old_exe)
+            try:
+                is_executable = os.access(old_exe, os.X_OK)
+            except OSError:
+                pass
+            else:
+                try:
+                    st = os.stat(old_exe)
+                except OSError:
+                    pass
+                else:
+                    if stat.S_ISDIR(st.st_mode):
+                        cwd = old_exe
+                        cmd = resolved_shell(self.opts)
+                    elif not is_executable:
+                        import shlex
+                        with suppress(OSError):
+                            with open(old_exe) as f:
+                                cmd = [kitty_exe(), '+hold']
+                                if f.read(2) == '#!':
+                                    line = f.read(4096).splitlines()[0]
+                                    cmd += shlex.split(line) + [old_exe]
+                                else:
+                                    cmd += [resolved_shell(self.opts)[0], cmd[0]]
         fenv: Dict[str, str] = {}
         if env:
             fenv.update(env)
         fenv['KITTY_WINDOW_ID'] = str(next_window_id())
-        if not is_macos and not is_wayland():
-            try:
-                fenv['WINDOWID'] = str(x11_window_id(self.os_window_id))
-            except Exception:
-                import traceback
-                traceback.print_exc()
+        pwid = platform_window_id(self.os_window_id)
+        if pwid is not None:
+            fenv['WINDOWID'] = str(pwid)
         ans = Child(cmd, cwd or self.cwd, self.opts, stdin, fenv, cwd_from, allow_remote_control=allow_remote_control)
         ans.fork()
         return ans
 
-    def _add_window(self, window: Window, location: Optional[str] = None) -> None:
-        self.active_window_idx = self.current_layout.add_window(self.windows, window, self.active_window_idx, location)
-        self.relayout_borders()
+    def _add_window(self, window: Window, location: Optional[str] = None, overlay_for: Optional[int] = None) -> None:
+        self.current_layout.add_window(self.windows, window, location, overlay_for)
+        self.mark_tab_bar_dirty()
+        self.relayout()
 
     def new_window(
         self,
@@ -354,13 +359,9 @@ class Tab:  # {{{
             self, child, self.opts, self.args, override_title=override_title,
             copy_colors_from=copy_colors_from, watchers=watchers
         )
-        if overlay_for is not None:
-            overlaid = next(w for w in self.windows if w.id == overlay_for)
-            window.overlay_for = overlay_for
-            overlaid.overlay_window_id = window.id
         # Must add child before laying out so that resize_pty succeeds
         get_boss().add_child(window)
-        self._add_window(window, location=location)
+        self._add_window(window, location=location, overlay_for=overlay_for)
         if marker:
             try:
                 window.set_marker(marker)
@@ -374,105 +375,54 @@ class Tab:  # {{{
             special_window: SpecialWindowInstance,
             location: Optional[str] = None,
             copy_colors_from: Optional[Window] = None,
-            allow_remote_control: bool = False
+            allow_remote_control: bool = False,
     ) -> Window:
         return self.new_window(
             use_shell=False, cmd=special_window.cmd, stdin=special_window.stdin,
             override_title=special_window.override_title,
             cwd_from=special_window.cwd_from, cwd=special_window.cwd, overlay_for=special_window.overlay_for,
             env=special_window.env, location=location, copy_colors_from=copy_colors_from,
-            allow_remote_control=allow_remote_control
+            allow_remote_control=allow_remote_control, watchers=special_window.watchers
         )
 
     def close_window(self) -> None:
-        if self.windows:
-            self.remove_window(self.windows[self.active_window_idx])
+        w = self.active_window
+        if w is not None:
+            self.remove_window(w)
 
     def close_other_windows_in_tab(self) -> None:
         if len(self.windows) > 1:
-            active_window = self.windows[self.active_window_idx]
+            active_window = self.active_window
             for window in tuple(self.windows):
                 if window is not active_window:
                     self.remove_window(window)
 
-    def previous_active_window_idx(self, num: int) -> Optional[int]:
-        try:
-            old_window_id = self.active_window_history[-num]
-        except IndexError:
-            return None
-        for idx, w in enumerate(self.windows):
-            if w.id == old_window_id:
-                return idx
-
     def remove_window(self, window: Window, destroy: bool = True) -> None:
-        idx = self.previous_active_window_idx(1)
-        next_window_id = None
-        if idx is not None:
-            next_window_id = self.windows[idx].id
-        active_window_idx = self.current_layout.remove_window(self.windows, window, self.active_window_idx)
+        self.windows.remove_window(window)
         if destroy:
             remove_window(self.os_window_id, self.id, window.id)
         else:
             detach_window(self.os_window_id, self.id, window.id)
-        if window.overlay_for is not None:
-            for idx, q in enumerate(self.windows):
-                if q.id == window.overlay_for:
-                    active_window_idx = idx
-                    next_window_id = q.id
-                    break
-        if next_window_id is None and len(self.windows) > active_window_idx:
-            next_window_id = self.windows[active_window_idx].id
-        if next_window_id is not None:
-            for idx, window in enumerate(self.windows):
-                if window.id == next_window_id:
-                    self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
-                    break
-            else:
-                self.active_window_idx = active_window_idx
-        else:
-            self.active_window_idx = active_window_idx
-        self.relayout_borders()
+        self.mark_tab_bar_dirty()
+        self.relayout()
         active_window = self.active_window
         if active_window:
             self.title_changed(active_window)
 
-    def detach_window(self, window: Window) -> Tuple[Optional[Window], Optional[Window]]:
-        underlaid_window: Optional[Window] = None
-        overlaid_window: Optional[Window] = window
-        if window.overlay_for:
-            for x in self.windows:
-                if x.id == window.overlay_for:
-                    underlaid_window = x
-                    break
-        elif window.overlay_window_id:
-            underlaid_window = window
-            overlaid_window = None
-            for x in self.windows:
-                if x.id == window.overlay_window_id:
-                    overlaid_window = x
-                    break
-        if overlaid_window is not None:
-            self.remove_window(overlaid_window, destroy=False)
-        if underlaid_window is not None:
-            self.remove_window(underlaid_window, destroy=False)
-        return underlaid_window, overlaid_window
+    def detach_window(self, window: Window) -> Tuple[Window, ...]:
+        windows = list(self.windows.windows_in_group_of(window))
+        windows.sort(key=attrgetter('id'))  # since ids increase in order of creation
+        for w in reversed(windows):
+            self.remove_window(w, destroy=False)
+        return tuple(windows)
 
     def attach_window(self, window: Window) -> None:
         window.change_tab(self)
         attach_window(self.os_window_id, self.id, window.id)
         self._add_window(window)
 
-    def set_active_window_idx(self, idx: int) -> None:
-        if idx != self.active_window_idx:
-            self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
-            self.relayout_borders()
-
-    def set_active_window(self, window: Window) -> None:
-        try:
-            idx = self.windows.index(window)
-        except ValueError:
-            return
-        self.set_active_window_idx(idx)
+    def set_active_window(self, x: Union[Window, int]) -> None:
+        self.windows.set_active_window_group_for(x)
 
     def get_nth_window(self, n: int) -> Optional[Window]:
         if self.windows:
@@ -481,17 +431,14 @@ class Tab:  # {{{
     def nth_window(self, num: int = 0) -> None:
         if self.windows:
             if num < 0:
-                idx = self.previous_active_window_idx(-num)
-                if idx is None:
-                    return
-                self.active_window_idx = self.current_layout.set_active_window(self.windows, idx)
+                self.windows.make_previous_group_active(-num)
             else:
-                self.active_window_idx = self.current_layout.activate_nth_window(self.windows, num)
+                self.current_layout.activate_nth_window(self.windows, num)
             self.relayout_borders()
 
     def _next_window(self, delta: int = 1) -> None:
         if len(self.windows) > 1:
-            self.active_window_idx = self.current_layout.next_window(self.windows, self.active_window_idx, delta)
+            self.current_layout.next_window(self.windows, delta)
             self.relayout_borders()
 
     def next_window(self) -> None:
@@ -502,19 +449,42 @@ class Tab:  # {{{
 
     prev_window = previous_window
 
-    def neighboring_window(self, which: str) -> None:
-        neighbors = self.current_layout.neighbors(self.windows, self.active_window_idx)
-        candidates = cast(Optional[Tuple[int, ...]], neighbors.get(which))
-        if candidates:
-            self.active_window_idx = self.current_layout.set_active_window(self.windows, candidates[0])
-            self.relayout_borders()
+    def most_recent_group(self, groups: Sequence[int]) -> Optional[int]:
+        groups_set = frozenset(groups)
 
-    def move_window(self, delta: int = 1) -> None:
-        self.active_window_idx = self.current_layout.move_window(self.windows, self.active_window_idx, delta)
-        self.relayout()
+        for window_id in reversed(self.windows.active_window_history):
+            group = self.windows.group_for_window(window_id)
+            if group and group.id in groups_set:
+                return group.id
+
+        if groups:
+            return groups[0]
+
+    def neighboring_group_id(self, which: EdgeLiteral) -> Optional[int]:
+        neighbors = self.current_layout.neighbors(self.windows)
+        candidates = neighbors.get(which)
+        if candidates:
+            return self.most_recent_group(candidates)
+
+    def neighboring_window(self, which: EdgeLiteral) -> None:
+        neighbor = self.neighboring_group_id(which)
+        if neighbor:
+            self.windows.set_active_group(neighbor)
+
+    def move_window(self, delta: Union[EdgeLiteral, int] = 1) -> None:
+        if isinstance(delta, int):
+            if self.current_layout.move_window(self.windows, delta):
+                self.relayout()
+        elif isinstance(delta, str):
+            neighbor = self.neighboring_group_id(delta)
+            if neighbor:
+                if self.current_layout.move_window_to_group(self.windows, neighbor):
+                    self.relayout()
 
     def move_window_to_top(self) -> None:
-        self.move_window(-self.active_window_idx)
+        n = self.windows.active_group_idx
+        if n > 0:
+            self.move_window(-n)
 
     def move_window_forward(self) -> None:
         self.move_window()
@@ -522,9 +492,9 @@ class Tab:  # {{{
     def move_window_backward(self) -> None:
         self.move_window(-1)
 
-    def list_windows(self, active_window: Optional[Window]) -> Generator[WindowDict, None, None]:
+    def list_windows(self, active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[WindowDict, None, None]:
         for w in self:
-            yield w.as_dict(is_focused=w is active_window)
+            yield w.as_dict(is_focused=w is active_window, is_self=w is self_window)
 
     def matches(self, field: str, pat: Pattern) -> bool:
         if field == 'id':
@@ -546,7 +516,7 @@ class Tab:  # {{{
         evict_cached_layouts(self.id)
         for w in self.windows:
             w.destroy()
-        self.windows = deque()
+        self.windows = WindowList(self)
 
     def __repr__(self) -> str:
         return 'Tab(title={}, id={})'.format(self.name or self.title, hex(id(self)))
@@ -560,8 +530,10 @@ class Tab:  # {{{
 
 class TabManager:  # {{{
 
-    def __init__(self, os_window_id: int, opts: Options, args: CLIOptions, startup_session: Optional[SessionType] = None):
+    def __init__(self, os_window_id: int, opts: Options, args: CLIOptions, wm_class: str, wm_name: str, startup_session: Optional[SessionType] = None):
         self.os_window_id = os_window_id
+        self.wm_class = wm_class
+        self.wm_name = wm_name
         self.last_active_tab_id = None
         self.opts, self.args = opts, args
         self.tab_bar_hidden = self.opts.tab_bar_style == 'hidden'
@@ -640,9 +612,10 @@ class TabManager:  # {{{
     def update_tab_bar_data(self) -> None:
         self.tab_bar.update(self.tab_bar_data)
 
-    def update_dpi_based_sizes(self) -> None:
-        for tab in self.tabs:
-            tab.recalculate_sizes()
+    def title_changed(self, tab: Tab) -> None:
+        self.mark_tab_bar_dirty()
+        if tab is self.active_tab:
+            sync_os_window_title(self.os_window_id)
 
     def resize(self, only_tabs: bool = False) -> None:
         if not only_tabs:
@@ -692,17 +665,26 @@ class TabManager:  # {{{
     def __len__(self) -> int:
         return len(self.tabs)
 
-    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window]) -> Generator[TabDict, None, None]:
+    def list_tabs(self, active_tab: Optional[Tab], active_window: Optional[Window], self_window: Optional[Window] = None) -> Generator[TabDict, None, None]:
         for tab in self:
             yield {
                 'id': tab.id,
                 'is_focused': tab is active_tab,
                 'title': tab.name or tab.title,
                 'layout': str(tab.current_layout.name),
-                'windows': list(tab.list_windows(active_window)),
-                'active_window_history': list(tab.active_window_history),
+                'layout_state': tab.current_layout.layout_state(),
+                'windows': list(tab.list_windows(active_window, self_window)),
+                'active_window_history': list(tab.windows.active_window_history),
             }
 
+    def serialize_state(self) -> Dict[str, Any]:
+        return {
+            'version': 1,
+            'id': self.os_window_id,
+            'tabs': [tab.serialize_state() for tab in self],
+            'active_tab_idx': self.active_tab_idx,
+        }
+
     @property
     def active_tab(self) -> Optional[Tab]:
         return self.tabs[self.active_tab_idx] if self.tabs else None
@@ -713,6 +695,14 @@ class TabManager:  # {{{
         if t is not None:
             return t.active_window
 
+    @property
+    def number_of_windows(self) -> int:
+        count = 0
+        for tab in self:
+            for window in tab:
+                count += 1
+        return count
+
     def tab_for_id(self, tab_id: int) -> Optional[Tab]:
         for t in self.tabs:
             if t.id == tab_id:
@@ -761,27 +751,34 @@ class TabManager:  # {{{
 
     def remove(self, tab: Tab) -> None:
         self._remove_tab(tab)
-        next_active_tab = -1
+        try:
+            active_tab_needs_to_change = self.active_tab is None or self.active_tab is tab
+        except IndexError:
+            active_tab_needs_to_change = True
         while True:
             try:
                 self.active_tab_history.remove(tab.id)
             except ValueError:
                 break
 
-        if self.opts.tab_switch_strategy == 'previous':
-            while self.active_tab_history and next_active_tab < 0:
-                tab_id = self.active_tab_history.pop()
-                for idx, qtab in enumerate(self.tabs):
-                    if qtab.id == tab_id:
-                        next_active_tab = idx
-                        break
-        elif self.opts.tab_switch_strategy == 'left':
-            next_active_tab = max(0, self.active_tab_idx - 1)
+        if active_tab_needs_to_change:
+            next_active_tab = -1
+            if self.opts.tab_switch_strategy == 'previous':
+                while self.active_tab_history and next_active_tab < 0:
+                    tab_id = self.active_tab_history.pop()
+                    for idx, qtab in enumerate(self.tabs):
+                        if qtab.id == tab_id:
+                            next_active_tab = idx
+                            break
+            elif self.opts.tab_switch_strategy == 'left':
+                next_active_tab = max(0, self.active_tab_idx - 1)
+            elif self.opts.tab_switch_strategy == 'right':
+                next_active_tab = min(self.active_tab_idx, len(self.tabs) - 1)
 
-        if next_active_tab < 0:
-            next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
+            if next_active_tab < 0:
+                next_active_tab = max(0, min(self.active_tab_idx, len(self.tabs) - 1))
 
-        self._set_active_tab(next_active_tab)
+            self._set_active_tab(next_active_tab)
         self.mark_tab_bar_dirty()
         tab.destroy()
 
@@ -792,16 +789,25 @@ class TabManager:  # {{{
         for t in self.tabs:
             title = (t.name or t.title or appname).strip()
             needs_attention = False
+            has_activity_since_last_focus = False
             for w in t:
                 if w.needs_attention:
                     needs_attention = True
-                    break
-            ans.append(TabBarData(title, t is at, needs_attention))
+                if w.has_activity_since_last_focus:
+                    has_activity_since_last_focus = True
+            ans.append(TabBarData(
+                title, t is at, needs_attention,
+                len(t), t.current_layout.name or '',
+                has_activity_since_last_focus
+            ))
         return ans
 
-    def activate_tab_at(self, x: int) -> None:
+    def activate_tab_at(self, x: int, is_double: bool = False) -> None:
         i = self.tab_bar.tab_at(x)
-        if i is not None:
+        if i is None:
+            if is_double:
+                self.new_tab()
+        else:
             self.set_active_tab_idx(i)
 
     @property
