@@ -8,6 +8,8 @@ import sys
 import weakref
 from collections import deque
 from enum import IntEnum
+from functools import partial
+from gettext import gettext as _
 from itertools import chain
 from typing import (
     Any, Callable, Deque, Dict, Iterable, List, Optional, Pattern, Sequence,
@@ -17,25 +19,28 @@ from typing import (
 from .child import ProcessDesc
 from .cli_stub import CLIOptions
 from .config import build_ansi_color_table
-from .constants import ScreenGeometry, WindowGeometry, appname, wakeup
+from .constants import appname, wakeup
 from .fast_data_types import (
     BGIMAGE_PROGRAM, BLIT_PROGRAM, CELL_BG_PROGRAM, CELL_FG_PROGRAM,
-    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, CSI, DCS, DECORATION, DIM,
+    CELL_PROGRAM, CELL_SPECIAL_PROGRAM, DCS, DECORATION, DIM, GLFW_MOD_CONTROL,
     GRAPHICS_ALPHA_MASK_PROGRAM, GRAPHICS_PREMULT_PROGRAM, GRAPHICS_PROGRAM,
     MARK, MARK_MASK, OSC, REVERSE, SCROLL_FULL, SCROLL_LINE, SCROLL_PAGE,
-    STRIKETHROUGH, TINT_PROGRAM, Screen, add_window, cell_size_for_window,
-    compile_program, get_boss, get_clipboard_string, init_cell_program,
-    set_clipboard_string, set_titlebar_color, set_window_render_data,
+    STRIKETHROUGH, TINT_PROGRAM, KeyEvent, Screen, add_timer, add_window,
+    cell_size_for_window, compile_program, encode_key_for_tty, get_boss,
+    get_clipboard_string, init_cell_program, pt_to_px, set_clipboard_string,
+    set_titlebar_color, set_window_padding, set_window_render_data,
     update_window_title, update_window_visibility, viewport_for_window
 )
-from .keys import defines, extended_key_event, keyboard_mode_name
+from .keys import keyboard_mode_name
+from .notify import NotificationCommand, handle_notification_cmd
 from .options_stub import Options
 from .rgb import to_color
 from .terminfo import get_capabilities
-from .typing import BossType, ChildType, TabType, TypedDict
+from .types import ScreenGeometry, WindowGeometry
+from .typing import BossType, ChildType, EdgeLiteral, TabType, TypedDict
 from .utils import (
-    color_as_int, get_primary_selection, load_shaders, open_cmd, open_url,
-    parse_color_set, read_shell_environment, sanitize_title,
+    color_as_int, get_primary_selection, load_shaders, log_error, open_cmd,
+    open_url, parse_color_set, read_shell_environment, sanitize_title,
     set_primary_selection
 )
 
@@ -51,6 +56,9 @@ class WindowDict(TypedDict):
     cmdline: List[str]
     env: Dict[str, str]
     foreground_processes: List[ProcessDesc]
+    is_self: bool
+    lines: int
+    columns: int
 
 
 class PipeData(TypedDict):
@@ -85,9 +93,48 @@ class Watcher:
 
 class Watchers:
 
+    on_resize: List[Watcher]
+    on_close: List[Watcher]
+    on_focus_change: List[Watcher]
+
     def __init__(self) -> None:
-        self.on_resize: List[Watcher] = []
-        self.on_close: List[Watcher] = []
+        self.on_resize = []
+        self.on_close = []
+        self.on_focus_change = []
+
+    def add(self, others: 'Watchers') -> None:
+        def merge(base: List[Watcher], other: List[Watcher]) -> None:
+            for x in other:
+                if x not in base:
+                    base.append(x)
+        merge(self.on_resize, others.on_resize)
+        merge(self.on_close, others.on_close)
+        merge(self.on_focus_change, others.on_focus_change)
+
+    def clear(self) -> None:
+        del self.on_close[:], self.on_resize[:], self.on_focus_change[:]
+
+    def copy(self) -> 'Watchers':
+        ans = Watchers()
+        ans.on_close = self.on_close[:]
+        ans.on_resize = self.on_resize[:]
+        ans.on_focus_change = self.on_focus_change
+        return ans
+
+    @property
+    def has_watchers(self) -> bool:
+        return bool(self.on_close or self.on_resize or self.on_focus_change)
+
+
+def call_watchers(windowref: Callable[[], Optional['Window']], which: str, data: Dict[str, Any]) -> None:
+
+    def callback(timer_id: Optional[int]) -> None:
+        w = windowref()
+        if w is not None:
+            watchers: List[Watcher] = getattr(w.watchers, which)
+            w.call_watchers(watchers, data)
+
+    add_timer(callback, 0, False)
 
 
 def calculate_gl_geometry(window_geometry: WindowGeometry, viewport_width: int, viewport_height: int, cell_width: int, cell_height: int) -> ScreenGeometry:
@@ -99,6 +146,38 @@ def calculate_gl_geometry(window_geometry: WindowGeometry, viewport_width: int,
     return ScreenGeometry(xstart, ystart, window_geometry.xnum, window_geometry.ynum, dx, dy)
 
 
+def as_text(
+    screen: Screen,
+    as_ansi: bool = False,
+    add_history: bool = False,
+    add_wrap_markers: bool = False,
+    alternate_screen: bool = False
+) -> str:
+    lines: List[str] = []
+    add_history = add_history and not (screen.is_using_alternate_linebuf() ^ alternate_screen)
+    if alternate_screen:
+        f = screen.as_text_alternate
+    else:
+        f = screen.as_text_non_visual if add_history else screen.as_text
+    f(lines.append, as_ansi, add_wrap_markers)
+    if add_history:
+        h: List[str] = []
+        pht = screen.historybuf.pagerhist_as_text()
+        if pht:
+            h.append(pht)
+        if h and (not as_ansi or not add_wrap_markers):
+            sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
+            h = list(map(sanitizer, h))
+        screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
+        if h:
+            if not screen.linebuf.is_continued(0):
+                h[-1] += '\n'
+            if as_ansi:
+                h[-1] += '\x1b[m'
+        return ''.join(chain(h, lines))
+    return ''.join(lines)
+
+
 class LoadShaderPrograms:
 
     use_selection_fg = True
@@ -166,7 +245,7 @@ def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str
     pat = getattr(text_sanitizer, 'pat', None)
     if pat is None:
         import re
-        pat = re.compile(r'\033\[.+?m')
+        pat = re.compile('\033\\[.*?m')
         setattr(text_sanitizer, 'pat', pat)
 
     ansi, wrap_markers = not as_ansi, not add_wrap_markers
@@ -185,6 +264,25 @@ def text_sanitizer(as_ansi: bool, add_wrap_markers: bool) -> Callable[[str], str
     return remove_wrap_markers
 
 
+class EdgeWidths:
+    left: Optional[float]
+    top: Optional[float]
+    right: Optional[float]
+    bottom: Optional[float]
+
+    def __init__(self, serialized: Optional[Dict[str, Optional[float]]] = None):
+        if serialized is not None:
+            self.left = serialized['left']
+            self.right = serialized['right']
+            self.top = serialized['top']
+            self.bottom = serialized['bottom']
+        else:
+            self.left = self.top = self.right = self.bottom = None
+
+    def serialize(self) -> Dict[str, Optional[float]]:
+        return {'left': self.left, 'right': self.right, 'top': self.top, 'bottom': self.bottom}
+
+
 class Window:
 
     def __init__(
@@ -198,19 +296,21 @@ class Window:
         watchers: Optional[Watchers] = None
     ):
         self.watchers = watchers or Watchers()
+        self.prev_osc99_cmd = NotificationCommand()
         self.action_on_close: Optional[Callable] = None
         self.action_on_removal: Optional[Callable] = None
         self.current_marker_spec: Optional[Tuple[str, Union[str, Tuple[Tuple[int, str], ...]]]] = None
         self.pty_resized_once = False
+        self.last_reported_pty_size = (-1, -1, -1, -1)
         self.needs_attention = False
         self.override_title = override_title
-        self.overlay_window_id: Optional[int] = None
-        self.overlay_for: Optional[int] = None
         self.default_title = os.path.basename(child.argv[0] or appname)
         self.child_title = self.default_title
         self.title_stack: Deque[str] = deque(maxlen=10)
         self.allow_remote_control = child.allow_remote_control
-        self.id = add_window(tab.os_window_id, tab.id, self.title)
+        self.id: int = add_window(tab.os_window_id, tab.id, self.title)
+        self.margin = EdgeWidths()
+        self.padding = EdgeWidths()
         if not self.id:
             raise Exception('No tab with id: {} in OS Window: {} was found, or the window counter wrapped'.format(tab.id, tab.os_window_id))
         self.tab_id = tab.id
@@ -218,31 +318,72 @@ class Window:
         self.tabref: Callable[[], Optional[TabType]] = weakref.ref(tab)
         self.clipboard_control_buffers = {'p': '', 'c': ''}
         self.destroyed = False
-        self.geometry = WindowGeometry(0, 0, 0, 0, 0, 0)
+        self.geometry: WindowGeometry = WindowGeometry(0, 0, 0, 0, 0, 0)
         self.needs_layout = True
-        self.is_visible_in_layout = True
+        self.is_visible_in_layout: bool = True
         self.child, self.opts = child, opts
         cell_width, cell_height = cell_size_for_window(self.os_window_id)
-        self.screen = Screen(self, 24, 80, opts.scrollback_lines, cell_width, cell_height, self.id)
+        self.screen: Screen = Screen(self, 24, 80, opts.scrollback_lines, cell_width, cell_height, self.id)
         if copy_colors_from is not None:
             self.screen.copy_colors_from(copy_colors_from.screen)
         else:
             setup_colors(self.screen, opts)
 
+    def on_dpi_change(self, font_sz: float) -> None:
+        self.update_effective_padding()
+
     def change_tab(self, tab: TabType) -> None:
         self.tab_id = tab.id
         self.os_window_id = tab.os_window_id
         self.tabref = weakref.ref(tab)
 
+    def effective_margin(self, edge: EdgeLiteral, is_single_window: bool = False) -> int:
+        q = getattr(self.margin, edge)
+        if q is not None:
+            return pt_to_px(q, self.os_window_id)
+        if is_single_window:
+            q = getattr(self.opts.single_window_margin_width, edge)
+            if q > -0.1:
+                return pt_to_px(q, self.os_window_id)
+        q = getattr(self.opts.window_margin_width, edge)
+        return pt_to_px(q, self.os_window_id)
+
+    def effective_padding(self, edge: EdgeLiteral) -> int:
+        q = getattr(self.padding, edge)
+        if q is not None:
+            return pt_to_px(q, self.os_window_id)
+        q = getattr(self.opts.window_padding_width, edge)
+        return pt_to_px(q, self.os_window_id)
+
+    def update_effective_padding(self) -> None:
+        set_window_padding(
+            self.os_window_id, self.tab_id, self.id,
+            self.effective_padding('left'), self.effective_padding('top'),
+            self.effective_padding('right'), self.effective_padding('bottom'))
+
+    def patch_edge_width(self, which: str, edge: EdgeLiteral, val: Optional[float]) -> None:
+        q = self.padding if which == 'padding' else self.margin
+        setattr(q, edge, val)
+        if q is self.padding:
+            self.update_effective_padding()
+
+    def effective_border(self) -> int:
+        val, unit = self.opts.window_border_width
+        if unit == 'pt':
+            val = max(1 if val > 0 else 0, pt_to_px(val, self.os_window_id))
+        else:
+            val = round(val)
+        return int(val)
+
     @property
     def title(self) -> str:
         return self.override_title or self.child_title
 
     def __repr__(self) -> str:
-        return 'Window(title={}, id={}, overlay_for={}, overlay_window_id={})'.format(
-                self.title, self.id, self.overlay_for, self.overlay_window_id)
+        return 'Window(title={}, id={})'.format(
+                self.title, self.id)
 
-    def as_dict(self, is_focused: bool = False) -> WindowDict:
+    def as_dict(self, is_focused: bool = False, is_self: bool = False) -> WindowDict:
         return dict(
             id=self.id,
             is_focused=is_focused,
@@ -251,9 +392,28 @@ class Window:
             cwd=self.child.current_cwd or self.child.cwd,
             cmdline=self.child.cmdline,
             env=self.child.environ,
-            foreground_processes=self.child.foreground_processes
+            foreground_processes=self.child.foreground_processes,
+            is_self=is_self,
+            lines=self.screen.lines,
+            columns=self.screen.columns,
         )
 
+    def serialize_state(self) -> Dict[str, Any]:
+        return {
+            'version': 1,
+            'id': self.id,
+            'child_title': self.child_title,
+            'override_title': self.override_title,
+            'default_title': self.default_title,
+            'title_stack': list(self.title_stack),
+            'allow_remote_control': self.allow_remote_control,
+            'cwd': self.child.current_cwd or self.child.cwd,
+            'env': self.child.environ,
+            'cmdline': self.child.cmdline,
+            'margin': self.margin.serialize(),
+            'padding': self.padding.serialize(),
+        }
+
     @property
     def current_colors(self) -> Dict:
         return self.screen.color_profile.as_dict()
@@ -271,7 +431,7 @@ class Window:
             return False
         assert not isinstance(pat, tuple)
 
-        if field == 'id':
+        if field in ('id', 'window_id'):
             return True if pat.pattern == str(self.id) else False
         if field == 'pid':
             return True if pat.pattern == str(self.child.pid) else False
@@ -286,11 +446,11 @@ class Window:
             return False
         return False
 
-    def set_visible_in_layout(self, window_idx: int, val: bool) -> None:
+    def set_visible_in_layout(self, val: bool) -> None:
         val = bool(val)
         if val is not self.is_visible_in_layout:
             self.is_visible_in_layout = val
-            update_window_visibility(self.os_window_id, self.tab_id, self.id, window_idx, val)
+            update_window_visibility(self.os_window_id, self.tab_id, self.id, val)
             if val:
                 self.refresh()
 
@@ -303,26 +463,29 @@ class Window:
         self.screen_geometry = sg = calculate_gl_geometry(window_geometry, vw, vh, cw, ch)
         return sg
 
-    def set_geometry(self, window_idx: int, new_geometry: WindowGeometry) -> None:
+    def set_geometry(self, new_geometry: WindowGeometry) -> None:
         if self.destroyed:
             return
         if self.needs_layout or new_geometry.xnum != self.screen.columns or new_geometry.ynum != self.screen.lines:
-            boss = get_boss()
             self.screen.resize(new_geometry.ynum, new_geometry.xnum)
-            current_pty_size = (
-                self.screen.lines, self.screen.columns,
-                max(0, new_geometry.right - new_geometry.left), max(0, new_geometry.bottom - new_geometry.top))
             sg = self.update_position(new_geometry)
             self.needs_layout = False
-            boss.child_monitor.resize_pty(self.id, *current_pty_size)
+            call_watchers(weakref.ref(self), 'on_resize', {'old_geometry': self.geometry, 'new_geometry': new_geometry})
+        else:
+            sg = self.update_position(new_geometry)
+        current_pty_size = (
+            self.screen.lines, self.screen.columns,
+            max(0, new_geometry.right - new_geometry.left), max(0, new_geometry.bottom - new_geometry.top))
+        if current_pty_size != self.last_reported_pty_size:
+            get_boss().child_monitor.resize_pty(self.id, *current_pty_size)
             if not self.pty_resized_once:
                 self.pty_resized_once = True
                 self.child.mark_terminal_ready()
-            self.call_watchers(self.watchers.on_resize, {'old_geometry': self.geometry, 'new_geometry': new_geometry})
-        else:
-            sg = self.update_position(new_geometry)
+            self.last_reported_pty_size = current_pty_size
+
         self.geometry = g = new_geometry
-        set_window_render_data(self.os_window_id, self.tab_id, self.id, window_idx, sg.xstart, sg.ystart, sg.dx, sg.dy, self.screen, *g[:4])
+        set_window_render_data(self.os_window_id, self.tab_id, self.id, sg.xstart, sg.ystart, sg.dx, sg.dy, self.screen, *g[:4])
+        self.update_effective_padding()
 
     def contains(self, x: int, y: int) -> bool:
         g = self.geometry
@@ -344,7 +507,7 @@ class Window:
     def write_to_child(self, data: Union[str, bytes]) -> None:
         if data:
             if get_boss().child_monitor.needs_write(self.id, data) is not True:
-                print('Failed to write to child %d as it does not exist' % self.id, file=sys.stderr)
+                log_error(f'Failed to write to child {self.id} as it does not exist')
 
     def title_updated(self) -> None:
         update_window_title(self.os_window_id, self.tab_id, self.id, self.title)
@@ -358,18 +521,80 @@ class Window:
         self.override_title = title or None
         self.title_updated()
 
+    def desktop_notify(self, osc_code: int, raw_data: str) -> None:
+        cmd = handle_notification_cmd(osc_code, raw_data, self.id, self.prev_osc99_cmd)
+        if cmd is not None and osc_code == 99:
+            self.prev_osc99_cmd = cmd
+
     # screen callbacks {{{
     def use_utf8(self, on: bool) -> None:
         get_boss().child_monitor.set_iutf8_winid(self.id, on)
 
+    def open_url(self, url: str, hyperlink_id: int, cwd: Optional[str] = None) -> None:
+        if hyperlink_id:
+            if not self.opts.allow_hyperlinks:
+                return
+            from urllib.parse import unquote, urlparse, urlunparse
+            try:
+                purl = urlparse(url)
+            except Exception:
+                return
+            if (not purl.scheme or purl.scheme == 'file'):
+                if purl.netloc:
+                    from socket import gethostname
+                    try:
+                        hostname = gethostname()
+                    except Exception:
+                        hostname = ''
+                    remote_hostname = purl.netloc.partition(':')[0]
+                    if remote_hostname and remote_hostname != hostname and remote_hostname != 'localhost':
+                        self.handle_remote_file(purl.netloc, unquote(purl.path))
+                        return
+                    url = urlunparse(purl._replace(netloc=''))
+            if self.opts.allow_hyperlinks & 0b10:
+                from kittens.tui.operations import styled
+                get_boss()._run_kitten('ask', ['--type=choices', '--message', _(
+                    'What would you like to do with this URL:\n') +
+                    styled(unquote(url), fg='yellow'),
+                    '--choice=o:Open', '--choice=c:Copy to clipboard', '--choice=n;red:Nothing'
+                    ],
+                    window=self,
+                    custom_callback=partial(self.hyperlink_open_confirmed, url, cwd)
+                )
+                return
+        get_boss().open_url(url, cwd=cwd)
+
+    def hyperlink_open_confirmed(self, url: str, cwd: Optional[str], data: Dict[str, Any], *a: Any) -> None:
+        q = data['response']
+        if q == 'o':
+            get_boss().open_url(url, cwd=cwd)
+        elif q == 'c':
+            set_clipboard_string(url)
+
+    def handle_remote_file(self, netloc: str, remote_path: str) -> None:
+        from kittens.ssh.main import get_connection_data
+        args = self.child.foreground_cmdline
+        conn_data = get_connection_data(args)
+        if conn_data is None:
+            get_boss().show_error('Could not handle remote file', 'No SSH connection data found in: {args}')
+            return
+        get_boss().run_kitten(
+            'remote_file', '--hostname', netloc.partition(':')[0], '--path', remote_path,
+            '--ssh-connection-data', json.dumps(conn_data)
+        )
+
     def focus_changed(self, focused: bool) -> None:
+        if self.destroyed:
+            return
+        call_watchers(weakref.ref(self), 'on_focus_change', {'focused': focused})
+        self.screen.focus_changed(focused)
         if focused:
+            changed = self.needs_attention
             self.needs_attention = False
-            if self.screen.focus_tracking_enabled:
-                self.screen.send_escape_code_to_child(CSI, 'I')
-        else:
-            if self.screen.focus_tracking_enabled:
-                self.screen.send_escape_code_to_child(CSI, 'O')
+            if changed:
+                tab = self.tabref()
+                if tab is not None:
+                    tab.relayout_borders()
 
     def title_changed(self, new_title: Optional[str]) -> None:
         self.child_title = sanitize_title(new_title or self.default_title)
@@ -383,17 +608,24 @@ class Window:
     def is_active(self) -> bool:
         return get_boss().active_window is self
 
+    @property
+    def has_activity_since_last_focus(self) -> bool:
+        return self.screen.has_activity_since_last_focus()
+
     def on_bell(self) -> None:
         if self.opts.command_on_bell and self.opts.command_on_bell != ['none']:
-            import subprocess
             import shlex
+            import subprocess
             env = self.child.final_env
             env['KITTY_CHILD_CMDLINE'] = ' '.join(map(shlex.quote, self.child.cmdline))
             subprocess.Popen(self.opts.command_on_bell, env=env, cwd=self.child.foreground_cwd)
         if not self.is_active:
+            changed = not self.needs_attention
             self.needs_attention = True
             tab = self.tabref()
             if tab is not None:
+                if changed:
+                    tab.relayout_borders()
                 tab.on_bell(self)
 
     def change_titlebar_color(self) -> None:
@@ -433,6 +665,9 @@ class Window:
         b |= b << 8
         self.screen.send_escape_code_to_child(OSC, '{};rgb:{:04x}/{:04x}/{:04x}'.format(code, r, g, b))
 
+    def report_notification_activated(self, identifier: str) -> None:
+        self.screen.send_escape_code_to_child(OSC, f'99;i={identifier};')
+
     def set_dynamic_color(self, code: int, value: Union[str, bytes]) -> None:
         if isinstance(value, bytes):
             value = value.decode('utf-8')
@@ -476,7 +711,8 @@ class Window:
             self.refresh()
 
     def request_capabilities(self, q: str) -> None:
-        self.screen.send_escape_code_to_child(DCS, get_capabilities(q))
+        for result in get_capabilities(q, self.opts):
+            self.screen.send_escape_code_to_child(DCS, result)
 
     def handle_remote_cmd(self, cmd: str) -> None:
         get_boss().handle_remote_cmd(cmd, self)
@@ -549,7 +785,7 @@ class Window:
         lines = self.screen.text_for_selection()
         if self.opts.strip_trailing_spaces == 'always' or (
                 self.opts.strip_trailing_spaces == 'smart' and not self.screen.is_rectangle_select()):
-            return ''.join((l.rstrip() or '\n') for l in lines)
+            return ''.join((ln.rstrip() or '\n') for ln in lines)
         return ''.join(lines)
 
     def call_watchers(self, which: Iterable[Watcher], data: Dict[str, Any]) -> None:
@@ -576,27 +812,7 @@ class Window:
         add_wrap_markers: bool = False,
         alternate_screen: bool = False
     ) -> str:
-        lines: List[str] = []
-        add_history = add_history and not (self.screen.is_using_alternate_linebuf() ^ alternate_screen)
-        if alternate_screen:
-            f = self.screen.as_text_alternate
-        else:
-            f = self.screen.as_text_non_visual if add_history else self.screen.as_text
-        f(lines.append, as_ansi, add_wrap_markers)
-        if add_history:
-            h: List[str] = []
-            self.screen.historybuf.pagerhist_as_text(h.append)
-            if h and (not as_ansi or not add_wrap_markers):
-                sanitizer = text_sanitizer(as_ansi, add_wrap_markers)
-                h = list(map(sanitizer, h))
-            self.screen.historybuf.as_text(h.append, as_ansi, add_wrap_markers)
-            if h:
-                if not self.screen.linebuf.is_continued(0):
-                    h[-1] += '\n'
-                if as_ansi:
-                    h[-1] += '\x1b[m'
-            return ''.join(chain(h, lines))
-        return ''.join(lines)
+        return as_text(self.screen, as_ansi, add_history, add_wrap_markers, alternate_screen)
 
     @property
     def cwd_of_child(self) -> Optional[str]:
@@ -623,7 +839,14 @@ class Window:
     def show_scrollback(self) -> None:
         text = self.as_text(as_ansi=True, add_history=True, add_wrap_markers=True)
         data = self.pipe_data(text, has_wrap_markers=True)
-        cmd = [x.replace('INPUT_LINE_NUMBER', str(data['input_line_number'])) for x in self.opts.scrollback_pager]
+
+        def prepare_arg(x: str) -> str:
+            x = x.replace('INPUT_LINE_NUMBER', str(data['input_line_number']))
+            x = x.replace('CURSOR_LINE', str(data['cursor_y']))
+            x = x.replace('CURSOR_COLUMN', str(data['cursor_x']))
+            return x
+
+        cmd = list(map(prepare_arg, self.opts.scrollback_pager))
         if not os.path.isabs(cmd[0]):
             import shutil
             exe = shutil.which(cmd[0])
@@ -665,14 +888,20 @@ class Window:
         if text:
             set_clipboard_string(text)
 
+    def encoded_key(self, key_event: KeyEvent) -> bytes:
+        return encode_key_for_tty(
+            key=key_event.key, shifted_key=key_event.shifted_key, alternate_key=key_event.alternate_key,
+            mods=key_event.mods, action=key_event.action, text=key_event.text,
+            key_encoding_flags=self.screen.current_key_encoding_flags(),
+            cursor_key_mode=self.screen.cursor_key_mode,
+        ).encode('ascii')
+
     def copy_or_interrupt(self) -> None:
         text = self.text_for_selection()
         if text:
             set_clipboard_string(text)
         else:
-            mode = keyboard_mode_name(self.screen)
-            data = extended_key_event(defines.GLFW_KEY_C, defines.GLFW_MOD_CONTROL, defines.GLFW_PRESS) if mode == 'kitty' else b'\x03'
-            self.write_to_child(data)
+            self.write_to_child(self.encoded_key(KeyEvent(key=ord('c'), mods=GLFW_MOD_CONTROL)))
 
     def copy_and_clear_or_interrupt(self) -> None:
         self.copy_or_interrupt()
@@ -721,7 +950,7 @@ class Window:
         self.current_marker_spec = key
 
     def set_marker(self, spec: Union[str, Sequence[str]]) -> None:
-        from .config import toggle_marker, parse_marker_spec
+        from .config import parse_marker_spec, toggle_marker
         from .marks import marker_from_spec
         if isinstance(spec, str):
             func, (ftype, spec_, flags) = toggle_marker('toggle_marker', spec)
@@ -738,4 +967,10 @@ class Window:
 
     def scroll_to_mark(self, prev: bool = True, mark: int = 0) -> None:
         self.screen.scroll_to_next_mark(mark, prev)
+
+    def signal_child(self, *signals: int) -> None:
+        pid = self.child.pid_for_cwd
+        if pid is not None:
+            for sig in signals:
+                os.kill(pid, sig)
     # }}}
